# solve the Poisson equation -Delta u = f
# with Dirichlet boundary condition u = 0

from ngsolve import *
from ngsolve.krylovspace import CG
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
import ngsolve
from xfem import *
from xfem.cutmg import * #MultiGridCL, CutFemSmoother, LinearMGIterator
from xfem.lsetcurv import *


ngsglobals.msg_level = 1

# generate a triangular mesh of mesh-size 0.2
square = SplineGeometry()
square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
mesh = Mesh (square.GenerateMesh(maxh=0.5, quad_dominated=False))

mu = [1e-3, 1]
order = 2
kappac="harmonic"
gamma_stab = 10


rhsscal = -4*mu[0]*mu[1] 
coef_f = [ rhsscal, rhsscal ]
distsq = x*x + y*y - 1
solution = [ mu[1]* distsq, mu[0] * distsq ]

levelset = ( sqrt(x*x + y*y) - 1 )
subdivlvl = 0

def main():
    print('hallo')
    
    nref = 3

    if( order > 1):
        mgprol=P2CutProlongation
    else:
        mgprol=P1Prolongation

    params =    {  
                    "order"     : order,
                    "mu"        : mu,
                    "gamma_stab": gamma_stab,
                    "kappa"     : "harmonic" # hansbo, highorder
                }

    for i in range(nref+1):
        if i==0:
            #lset
            #lsetp1 = GridFunction(H1(mesh, order = 1))
            #InterpolateToP1(levelset,lsetp1)
            #ci = CutInfo(mesh, lsetp1)
            (lsetmeshadap, deformation, lsetp1 ) = CreateLsetMeshAdapt(order)
            ci = CutInfo(mesh,lsetp1)
            #cut space
            Vh = H1(mesh, order=order, dirichlet=[1,2,3,4])
            VhNeg = Compress( Vh, active_dofs = GetDofsOfElements(Vh,ci.GetElementsOfType(HASNEG)))
            VhPos = Compress( Vh, active_dofs = GetDofsOfElements(Vh,ci.GetElementsOfType(HASPOS)))
            CutVh = FESpace( [VhNeg, VhPos], flags={"dgjumps": True} )
        else:
            print('before refine')
            mesh.Refine()
            print('after refine')

            #lsetp1.space.Update()
            #lsetp1.Update()
            #InterpolateToP1(levelset,lsetp1)
            #ci = CutInfo(mesh,lsetp1)
            print('before mesh adapt')
            (lsetmeshadap, deformation, lsetp1 ) = CreateLsetMeshAdapt(order)
            print('before cutinfo')
            ci = CutInfo(mesh,lsetp1)
            print('after mesh adapt')

            Vh = CutVh.components[0].GetBaseSpace()
            Vh.Update()
            print('update space')
            CutVh.components[0].SetActiveDofs( GetDofsOfElements(Vh,ci.GetElementsOfType(HASNEG)))
            CutVh.components[1].SetActiveDofs( GetDofsOfElements(Vh,ci.GetElementsOfType(HASPOS)))
            CutVh.Update()
            

        print('assemble problem')
        a, f = AssembleCutPoisson( CutVh=CutVh, ci=ci, lsetp1=lsetp1, highorder=(order>1), deformation=deformation )

        if i==0:
            mgiter = LinearMGIterator(a=a, ci=ci, lsetp1=lsetp1, mesh=mesh,
                                  nu=3,maxit=20, printinfo=True, ProlType=mgprol,
                                  ifsolver="cg" )
        else:
            mgiter.Update( a, CutVh, ci )



    #Vh = H1(mesh, order=order, dirichlet=[1,2,3,4])
    #(lsetmeshadap, deformation, lsetp1 ) = CreateLsetMeshAdapt(order)
    #ci = CutInfo(mesh,lsetp1)

    #HoCutFes = CutFESpace( Vh, ci, flags={"dgjumps": True} )    

    #a, f = AssembleCutPoisson( CutVh=HoCutFes, ci=ci, lsetp1=lsetp1, highorder=True, deformation=deformation )

    
    # tgiter = P2TwoGridCL( a=a, linmgiter=mgiter, fes=HoCutFes, ci=ci, nu=3, maxit=50,
    #                         mesh=mesh, ifsolver="cg", patchtype='edge' )


    gfu = GridFunction( CutVh )
    gfu.components[1].Set( solution[1], BND )
    f.vec.data -= a.mat * gfu.vec
    update = gfu.vec.CreateVector()

    update = mgiter * f.vec

    gfu.vec.data += update

    #udraw = tgiter.iterate(nu=3,tol=1e-8) #udraw = mgiter.iterate(rhs)    

    if( order > 1):
        mesh.SetDeformation(deformation)

    # Computation of L2 error:
    err_sqr_coefs = [(gfu.components[i]-solution[i])*(gfu.components[i]-solution[i]) for i in [0,1] ]
    lset_doms = LsetDoms( levelset, lsetp1, subdivlvl )
    # input('w2')
    l2error = sqrt( sum( [Integrate(lset_doms[i], cf=err_sqr_coefs[i], mesh=mesh, order=2*order, heapsize=1000000) for i in [0,1] ]))   
    #l2error = sqrt( sum( [Integrate(lset_doms[i], cf=err_sqr_coefs[i], mesh=mesh) for i in [0,1] ]))   
    print("L2 error : ",l2error)

    input('draw?\n')

    # if( nref > 2):
    #     print('nref too large')
    #     return
    #     print('test')

    print('drawing')
    # pretty printing    
    ucoef = IfPos( lsetp1, gfu.components[1], gfu.components[0] )
    if order>1 :        
        mesh.UnsetDeformation()
    Draw (ucoef,mesh,'u')



def LsetDoms( levelset, lsetp1, subdivlvl ):
    if (subdivlvl == 0):
        lsetint = lsetp1
    else:
        lsetint = levelset

    lset_neg = { "levelset" : lsetint, "domain_type" : NEG, "subdivlvl" : subdivlvl}
    lset_pos = { "levelset" : lsetint, "domain_type" : POS, "subdivlvl" : subdivlvl}
    lset_if  = { "levelset" : lsetint, "domain_type" : IF , "subdivlvl" : subdivlvl}

    lset_dom = [ lset_neg, lset_pos,  lset_if ]

    return lset_dom

def KappaLambda(ci,order):
    #kappac  = params["kappa"]
    #mu      = params["mu"]
    #order   = params["order"]

    if kappac == "harmonic":
        kappa = [ mu[1] / ( mu[0] + mu[1] ) , mu[0] /  ( mu[0] + mu[1] ) ]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 2*mu[0]*mu[1]/(mu[0] + mu[1]) * 10 * order * order
    elif kappac == "hansbo":
        kappa = [CutRatioGF(ci),1.0-CutRatioGF(ci)]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * (order+1) * order
    elif kappac == "highorder":
        kappaHO = IfPos( CutRatioGF(ci) -0.5, 1, 0 )
        kappa = [kappaHO,1.0-kappaHO]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * order * order

    return [ kappa, lambda_nitsche ]

def AssembleCutPoisson(**kwargs):
    ci = kwargs['ci']
    hasneg      = ci.GetElementsOfType(HASNEG)  # <- "hasneg": has (also) negative level set values
    haspos      = ci.GetElementsOfType(HASPOS)  # <- "haspos": has (also) positive level set values    
    hasif       =  BitArray(ci.GetElementsOfType(IF))
    ba_facets   = [GetFacetsWithNeighborTypes(mesh,a=hasneg,b=hasif),GetFacetsWithNeighborTypes(mesh,a=haspos,b=hasif)]


    lsetp1 = kwargs['lsetp1']
    n_lset  = 1.0/grad(lsetp1).Norm() * grad(lsetp1)
    h       = specialcf.mesh_size
    #params = kwargs['params']
    if( kwargs['highorder'] ):
        kappaorder = order
    else:
        kappaorder = 1

    [kappa, lambda_nitsche ] = KappaLambda(ci, kappaorder)    

    # expressions of test and trial functions (u and v are tuples):
    CutVh = kwargs['CutVh']
    (u,v) = CutVh.TnT()

    gradu = [grad(ui) for ui in u]
    gradv = [grad(vi) for vi in v]

    average_flux_u = sum([- kappa[i] * mu[i] * gradu[i] * n_lset for i in [0,1]])
    average_flux_v = sum([- kappa[i] * mu[i] * gradv[i] * n_lset for i in [0,1]])

    # integration domains
    lset_doms = LsetDoms( levelset, lsetp1, subdivlvl )
    
        # bilinear forms:
    a = BilinearForm(CutVh, symmetric = False)
    f = LinearForm(CutVh)

    #domain integrals:
    for i in [0,1] :
        a += SymbolicBFI( lset_doms[i], form = mu[i] * gradu[i] * gradv[i] )        

        #rhs
        f += SymbolicLFI( lset_doms[i], form = coef_f[i] * v[i] )

    # Nitsche integrals:
    a += SymbolicBFI(lset_doms[2], form = average_flux_u * (v[0]-v[1])
                                        + average_flux_v * (u[0]-u[1])
                                        + lambda_nitsche / h * (u[0]-u[1]) * (v[0]-v[1]) ) 

    for i in [0,1]:
        # Ghost penalty
        a += SymbolicFacetPatchBFI( form =  gamma_stab  * 
                                    mu[i] / h / h * 
                                    (u[i] - u[i].Other()) * (v[i] - v[i].Other()), 
                                    skeleton=False, definedonelements=ba_facets[i] )

    # # apply mesh adaptation 
    # if( doDeform ):   
    #     mesh.SetDeformation(deformation)

    if( kwargs['highorder'] ):
        print('setting deformation')
        mesh.SetDeformation( kwargs['deformation'] )

    # setting up matrix and vector
    a.Assemble()
    f.Assemble()

    if( kwargs['highorder'] ):
        print('setting deformation')
        mesh.UnsetDeformation()



    #updating with non-homogeneous boundary conditions
    #gfu = GridFunction( CutVh )
    #boundary lies in positive part
    #gfu.components[1].Set( solution[1], BND )
    #f.vec.data -= a.mat * gfu.vec

    return (a,f)
    

def CutFESpace( V, ci, flags=None ):    
    hasneg = ci.GetElementsOfType(HASNEG)  
    haspos = ci.GetElementsOfType(HASPOS)

    if isinstance(V,list):
        if len(V) != 2:
            raise ValueError("you need to provide a list of two FE spaces")
        else:
            return FESpace( [   Compress(V[0],active_dofs=GetDofsOfElements(V[0],hasneg) ), 
                                Compress(V[1],active_dofs=GetDofsOfElements(V[1],haspos) ) 
                            ], flags=flags
                          ) 
    else:
        return FESpace( [   Compress(V,active_dofs=GetDofsOfElements(V,hasneg) ), 
                            Compress(V,active_dofs=GetDofsOfElements(V,haspos) ) 
                        ], flags=flags
                      ) 



def CreateLsetMeshAdapt(order):
    print('call mesh adapt')
    lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=1000, discontinuous_qn=True)
    print('call calc deformation')
    deformation = lsetmeshadap.CalcDeformation(levelset)
    print('get lsetp1')
    lsetp1 = lsetmeshadap.lset_p1

    print('call return components')
    return (lsetmeshadap, deformation, lsetp1 )


if __name__ == '__main__':
    main()   



import pytest
from math import pi
from ngsolve import *
from netgen.geom2d import unit_square
from xfem import *

@pytest.mark.parametrize("order", [0,1,2,3,4,5])
def test_spacetime_set(order):
    ngmesh = unit_square.GenerateMesh(maxh=0.8, quad_dominated=False)
    mesh = Mesh (ngmesh)

    coef_told = Parameter(0)
    delta_t = 1
    tref = ReferenceTimeVariable()
    t = coef_told + delta_t*tref

    k_s = k_t = order
    fes1 = H1(mesh, order=k_s)

    tfe = ScalarTimeFE(k_t) 
    tfe_i = ScalarTimeFE(k_t, skip_first_node=True) # interior
    tfe_e = ScalarTimeFE(k_t, only_first_node=True) # exterior (inital values)

    st_fes = SpaceTimeFESpace(fes1,tfe, flags = {"dgjumps": True})
    st_fes_i = SpaceTimeFESpace(fes1,tfe_i, flags = {"dgjumps": True})
    st_fes_e = SpaceTimeFESpace(fes1,tfe_e, flags = {"dgjumps": True})

    gfu_slice = GridFunction(fes1)

    gfu = GridFunction(st_fes)
    gfu_i = GridFunction(st_fes_i)
    gfu_e = GridFunction(st_fes_e)

    for gf in [gfu,gfu_i,gfu_e]:
        V = gf.space
        gf.Set( 1 + t  )
        for i,that in enumerate(V.TimeFE_nodes()):
            if V.IsTimeNodeActive(i):
                val = 1 + that
            else:
                val = 0
            RestrictGFInTime(gf,that,gfu_slice)
            avg = Integrate(gfu_slice,mesh)
            assert(abs(avg-val) < 1e-12)

if __name__ == "__main__":
    test_spacetime_set(2)
    



from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

square = SplineGeometry()
square.AddRectangle((-4, -3), (4, 3), bc=1)
mesh = Mesh(square.GenerateMesh(maxh=2, quad_dominated=False))


level_sets = [x - 3, -x - y + 1, -x + y + 1,
              -x - 3, x - y + 1, x + y + 1 ]

level_sets_p1 = tuple(GridFunction(H1(mesh, order=1))
                      for i in range(len(level_sets)))
for i, lsetp1 in enumerate(level_sets_p1):
    InterpolateToP1(level_sets[i], lsetp1)
    Draw(lsetp1, mesh, "lsetp1_{}".format(i))

triangle = DomainTypeArray([(NEG, NEG, NEG, ANY, ANY, ANY),
                            (ANY, ANY, ANY, NEG, NEG, NEG)])

Draw(triangle.Indicator(level_sets_p1), mesh, "triangle",sd=6)
Draw(triangle.Boundary().IndicatorSmoothed(level_sets_p1), mesh, "bnd",sd=6)
Draw(triangle.Boundary().Boundary().IndicatorSmoothed(level_sets_p1), mesh, "cnr",sd=6)



"""
In this example we solve an unfitted Poisson problem similar to the one in
`fictdom.py`, however this time with the unfitted geometry being the
unit square. This example shall illustrate the functionality of ngsxfem to
solve PDE problems on geometries described via multiple level set functions.

PDE problem + Discretisation + Geometry + Implementation aspects:
-----------------------------------------------------------------
* As in fictdom.py except for the different geometry and its handling.

Used Features:
--------------
* Quadrature with respect to multiple level set functions., see the
  'mlset_pde' jupyter tutorial.

* MultiLevelsetCutInfo, see the 'mlset_basic' jupyter tutorial.

* DomainTypeArray convenience layer, see the 'mlset_basic' jupyter
  tutorial.

* Restricted BilinearForm, jupyter tutorial `basics`.

* Cut Differential Symbols, jupyter tutorials `intlset` and `cutfem`.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

ngsglobals.msg_level = 2

# -------------------------------- PARAMETERS ---------------------------------
# Domain corners
ll, ur = (-0.2, -0.2), (1.2, 1.2)
# Initial mesh diameter
initial_maxh = 0.4
# Number of mesh bisections
nref = 3
# Order of finite element space
k = 1

# Stabilization parameter for ghost-penalty
gamma_s = 0.5
# Stabilization parameter for Nitsche
gamma_n = 10

# ----------------------------------- MAIN ------------------------------------
# Set up the level sets, exact solution and right-hand side


def level_sets():
    return [-y, x - 1, y - 1, -x]


nr_ls = len(level_sets())
u_ex = 16 * x * (1 - x) * y * (1 - y)
grad_u_ex = (u_ex.Diff(x).Compile(), u_ex.Diff(y).Compile())
rhs = -(u_ex.Diff(x).Diff(x) + u_ex.Diff(y).Diff(y)).Compile()

# Geometry and mesh
geo = SplineGeometry()
geo.AddRectangle(ll, ur, bcs=("bottom", "right", "top", "left"))
ngmesh = geo.GenerateMesh(maxh=initial_maxh)
for i in range(nref):
    ngmesh.Refine()
mesh = Mesh(ngmesh)


# Level set and cut-information
P1 = H1(mesh, order=1)
lsetsp1 = tuple(GridFunction(P1) for i in range(nr_ls))
for i, lsetp1 in enumerate(lsetsp1):
    InterpolateToP1(level_sets()[i], lsetp1)
    Draw(lsetp1, mesh, "lsetp1_{}".format(i))

square = DomainTypeArray((NEG, NEG, NEG, NEG))
with TaskManager():
    square.Compress(lsetsp1)
    boundary = square.Boundary()
    boundary.Compress(lsetsp1)

mlci = MultiLevelsetCutInfo(mesh, lsetsp1)


# Element and degrees-of-freedom markers
els_if_singe = {dtt: BitArray(mesh.ne) for dtt in boundary}
facets_gp = BitArray(mesh.nedge)

hasneg = mlci.GetElementsWithContribution(square)

# Finite element space
Vhbase = H1(mesh, order=k, dgjumps=True)
Vh = Restrict(Vhbase, hasneg)
gfu = GridFunction(Vh)

hasif = mlci.GetElementsWithContribution(boundary)
Draw(BitArrayCF(hasif), mesh, "hasif")

for i, (dtt, els_bnd) in enumerate(els_if_singe.items()):
    els_bnd[:] = mlci.GetElementsWithContribution(dtt)
    Draw(BitArrayCF(els_bnd), mesh, "els_if_singe" + str(i))

facets_gp = GetFacetsWithNeighborTypes(mesh, a=hasneg, b=hasif,
                                       use_and=True)

els_gp = GetElementsWithNeighborFacets(mesh, facets_gp)
Draw(BitArrayCF(els_gp), mesh, "gp_elements")

# Bilinear and linear forms of the weak formulation
u, v = Vh.TnT()
h = specialcf.mesh_size
normals = square.GetOuterNormals(lsetsp1)

# Set up the integrator symbols
dx = dCut(lsetsp1, square, definedonelements=hasneg)
ds = {dtt: dCut(lsetsp1, dtt, definedonelements=els_if_singe[dtt])
      for dtt in boundary}
dw = dFacetPatch(definedonelements=facets_gp)

# Construct integrator
a = RestrictedBilinearForm(Vh, facet_restriction=facets_gp, check_unused=False)
a += InnerProduct(grad(u), grad(v)) * dx
for bnd, n in normals.items():
    a += -InnerProduct(grad(u) * n, v) * ds[bnd]
    a += -InnerProduct(grad(v) * n, u) * ds[bnd]
    a += (gamma_n * k * k / h) * InnerProduct(u, v) * ds[bnd]
a += gamma_s / (h**2) * (u - u.Other()) * (v - v.Other()) * dw

f = LinearForm(Vh)
f += rhs * v * dx


# Assemble and solve the linear system
f.Assemble()
a.Assemble()

gfu.vec.data = a.mat.Inverse(Vh.FreeDofs()) * f.vec

Draw(gfu, mesh, "uh")

# Post-processing
err_l2 = sqrt(Integrate((gfu - u_ex)**2 * dx.order(2 * k), mesh))
err_h1 = sqrt(Integrate((Grad(gfu) - grad_u_ex)**2 * dx.order(2 * (k - 1)),
                        mesh))

print("L2 error = {:1.5e}".format(err_l2), "H1 error = {:1.5e}".format(err_h1))



"""
Heat equation (fitted) with nonhomogeneous Dirichlet b.c. solved with a
DG-in-time space-time finite element discretisation

Domain:
-------
The domain of this examples is [0,1]x[0,1]x[0,1] (2D+time interval).

PDE problem:
------------
  u_t - (u_xx + u_yy) = f in  Omega=[0,1]^2
                    u = sin(pi * t) * cos(pi * x) * sin(pi * y) on dOmega
The r.h.s. term f is chosen according to a manufactured solution.

Discretisation:
---------------
* Discontinuous-in-time Galerkin space-time finite elements
* Weak imposition of boundary conditions using Nitsche's method

Implementational aspects:
-------------------------------------------------------
* A (sparse) direct solver is applied to solve the arising linear systems.

References:
-----------
All concepts that are used here are explained in the jupyter-tutorial
`spacetime.ipynb` (for the more involved case of moving domains).
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import unit_square
from ngsolve import *
from xfem import *
from math import pi
ngsglobals.msg_level = 1

# -------------------------------- PARAMETERS ---------------------------------
# Space finite element order
order = 2
# Time finite element order
k_t = order
# Final simulation time
tend = 0.5
# Time step
delta_t = 1 / 8
time_order = 2*k_t

# ----------------------------------- MAIN ------------------------------------
mesh = Mesh(unit_square.GenerateMesh(maxh=0.1, quad_dominated=False))
n = specialcf.normal(mesh.dim)
h = specialcf.mesh_size

V = H1(mesh, order=order, dirichlet=[], dgjumps=False)
# V = H1(mesh, order=order, dirichlet=".*")
tfe = ScalarTimeFE(k_t)
st_fes = tfe * V

# Fitted heat equation example
tnew = 0
told = Parameter(0)
t = told + delta_t * tref

u_exact = sin(pi * t) * cos(pi * x) * sin(pi * y)

coeff_f = u_exact.Diff(t) - (u_exact.Diff(x).Diff(x) + u_exact.Diff(y).Diff(y))
coeff_f = coeff_f.Compile()

gfu = GridFunction(st_fes)
u_last = CreateTimeRestrictedGF(gfu, 1)
u, v = st_fes.TnT()

dxt = delta_t * dxtref(mesh, time_order=time_order)
dxold = dmesh(mesh, tref=0)
dxnew = dmesh(mesh, tref=1)
# for Nitsche terms:
dst = delta_t * dxtref(mesh, time_order=time_order, skeleton=True, vb=BND)


def dt(u):
    return 1.0 / delta_t * dtref(u)


a = BilinearForm(st_fes, symmetric=False)
a += grad(u) * grad(v) * dxt
a += u * v * dxold
a += dt(u) * v * dxt

lam = 100
# Nitsche terms
a += (-1) * grad(u) * n * v * dst
a += (-1) * grad(v) * n * u * dst
a += lam * (1/h) * u * v * dst
a.Assemble()

f = LinearForm(st_fes)
f += coeff_f * v * dxt
f += u_last * v * dxold
# Nitsche terms
f += lam * (1/h) * u_exact * v * dst
f += -n * grad(v) * u_exact * dst
f.Assemble()

Draw(u_last, mesh, "u")

while tend - told.Get() > delta_t / 2:
    f.Assemble()
    gfu.vec.data = a.mat.Inverse(st_fes.FreeDofs(), "umfpack") * f.vec
    RestrictGFInTime(spacetime_gf=gfu, reference_time=1.0, space_gf=u_last)
    l2error = sqrt(Integrate((u_exact - gfu)**2 * dxnew, mesh))
    Redraw()
    told.Set(told.Get() + delta_t)
    print("\rt = {0:12.9f}, L2 error = {1:12.9e}".format(told.Get(), l2error))



# solve the Poisson equation -Delta u = f
# with Dirichlet boundary condition u = 0

from ngsolve import *
from ngsolve.krylovspace import CG
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from xfem import *
from xfem.lsetcurv import *


ngsglobals.msg_level = 1

# generate a triangular mesh of mesh-size 0.2
square = SplineGeometry()
square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
mesh = Mesh (square.GenerateMesh(maxh=0.5, quad_dominated=False))

mu = [1e-5, 1]
order = 2
kappac="harmonic"
gamma_stab = 10


rhsscal = -4*mu[0]*mu[1] 
coef_f = [ rhsscal, rhsscal ]
distsq = x*x + y*y - 1
solution = [ mu[1]* distsq, mu[0] * distsq ]

levelset = ( sqrt(x*x + y*y) - 1 )
subdivlvl = 0

def main():
    print('hallo')
    
    nref = 1

    params =    {  
                    "order"     : order,
                    "mu"        : mu,
                    "gamma_stab": gamma_stab,
                    "kappa"     : "harmonic" # hansbo, highorder
                }

    mgiter = LinearMGIterator(nref=nref,nu=1)    
    #lsetp1 = mgiter.getLset()            

    Vh = H1(mesh, order=order, dirichlet=[1,2,3,4])
    (lsetmeshadap, deformation, lsetp1 ) = CreateLsetMeshAdapt(order)
    ci = CutInfo(mesh,lsetp1)

    HoCutFes = CutFESpace( Vh, ci, flags={"dgjumps": True} )


    # VhNeg = Compress( Vh, active_dofs = GetActiveDof(mesh, HASNEG, order) )
    # VhPos = Compress( Vh, active_dofs = GetActiveDof(mesh, HASPOS, order) )

    # HoCutFes = FESpace( [VhNeg, VhPos], flags={"dgjumps": True} )

    

    a, f = AssembleCutPoisson( CutVh=HoCutFes, ci=ci, lsetp1=lsetp1, highorder=True, deformation=deformation )

    
    tgiter = HoTwoGridCL(a=a,f=f,mgiter=mgiter,fes=HoCutFes,ci=ci)

    #rhs = mgiter.getRhs()

    udraw = tgiter.iterate(nu=3,tol=1e-8) #udraw = mgiter.iterate(rhs)    

    # Computation of L2 error:
    err_sqr_coefs = [(udraw.components[i]-solution[i])*(udraw.components[i]-solution[i]) for i in [0,1] ]
    lset_doms = LsetDoms( levelset, lsetp1, subdivlvl )
    # input('w2')
    l2error = sqrt( sum( [Integrate(lset_doms[i], cf=err_sqr_coefs[i], mesh=mesh, order=2*order, heapsize=1000000) for i in [0,1] ]))   
    #l2error = sqrt( sum( [Integrate(lset_doms[i], cf=err_sqr_coefs[i], mesh=mesh) for i in [0,1] ]))   
    print("L2 error : ",l2error)

    input('draw?\n')

    # if( nref > 2):
    #     print('nref too large')
    #     return
    #     print('test')

    print('drawing')
    # pretty printing    
    ucoef = IfPos( lsetp1, udraw.components[1], udraw.components[0] )
    if order>1 :
        mesh.UnsetDeformation()
    Draw (ucoef,mesh,'u')

    input('finish? - seg fault\n')
    

def GetActiveDof(mesh,HASPOSNEG,order):
        Vh = H1(mesh, order = order, dirichlet=[], dgjumps = True)
        lsetp1 = GridFunction(Vh)
        InterpolateToP1(levelset,lsetp1)
        ci = CutInfo(mesh,lsetp1)
        return GetDofsOfElements(Vh,ci.GetElementsOfType(HASPOSNEG))


def LsetDoms( levelset, lsetp1, subdivlvl ):
    if (subdivlvl == 0):
        lsetint = lsetp1
    else:
        lsetint = levelset

    lset_neg = { "levelset" : lsetint, "domain_type" : NEG, "subdivlvl" : subdivlvl}
    lset_pos = { "levelset" : lsetint, "domain_type" : POS, "subdivlvl" : subdivlvl}
    lset_if  = { "levelset" : lsetint, "domain_type" : IF , "subdivlvl" : subdivlvl}

    lset_dom = [ lset_neg, lset_pos,  lset_if ]

    return lset_dom

def KappaLambda(ci,order):
    #kappac  = params["kappa"]
    #mu      = params["mu"]
    #order   = params["order"]

    if kappac == "harmonic":
        kappa = [ mu[1] / ( mu[0] + mu[1] ) , mu[0] /  ( mu[0] + mu[1] ) ]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 2*mu[0]*mu[1]/(mu[0] + mu[1]) * 10 * order * order
    elif kappac == "hansbo":
        kappa = [CutRatioGF(ci),1.0-CutRatioGF(ci)]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * (order+1) * order
    elif kappac == "highorder":
        kappaHO = IfPos( CutRatioGF(ci) -0.5, 1, 0 )
        kappa = [kappaHO,1.0-kappaHO]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * order * order

    return [ kappa, lambda_nitsche ]

def AssembleCutPoisson(**kwargs):
    ci = kwargs['ci']
    hasneg      = ci.GetElementsOfType(HASNEG)  # <- "hasneg": has (also) negative level set values
    haspos      = ci.GetElementsOfType(HASPOS)  # <- "haspos": has (also) positive level set values    
    hasif       =  BitArray(ci.GetElementsOfType(IF))
    ba_facets   = [GetFacetsWithNeighborTypes(mesh,a=hasneg,b=hasif),GetFacetsWithNeighborTypes(mesh,a=haspos,b=hasif)]


    lsetp1 = kwargs['lsetp1']
    n_lset  = 1.0/grad(lsetp1).Norm() * grad(lsetp1)
    h       = specialcf.mesh_size
    #params = kwargs['params']
    if( kwargs['highorder'] ):
        kappaorder = order
    else:
        kappaorder = 1

    [kappa, lambda_nitsche ] = KappaLambda(ci, kappaorder)    

    # expressions of test and trial functions (u and v are tuples):
    CutVh = kwargs['CutVh']
    (u,v) = CutVh.TnT()

    gradu = [grad(ui) for ui in u]
    gradv = [grad(vi) for vi in v]

    average_flux_u = sum([- kappa[i] * mu[i] * gradu[i] * n_lset for i in [0,1]])
    average_flux_v = sum([- kappa[i] * mu[i] * gradv[i] * n_lset for i in [0,1]])

    # integration domains
    lset_doms = LsetDoms( levelset, lsetp1, subdivlvl )
    
        # bilinear forms:
    a = BilinearForm(CutVh, symmetric = False)
    f = LinearForm(CutVh)

    #domain integrals:
    for i in [0,1] :
        a += SymbolicBFI( lset_doms[i], form = mu[i] * gradu[i] * gradv[i] )        

        #rhs
        f += SymbolicLFI( lset_doms[i], form = coef_f[i] * v[i] )

    # Nitsche integrals:
    a += SymbolicBFI(lset_doms[2], form = average_flux_u * (v[0]-v[1])
                                        + average_flux_v * (u[0]-u[1])
                                        + lambda_nitsche / h * (u[0]-u[1]) * (v[0]-v[1]) ) 

    for i in [0,1]:
        # Ghost penalty
        a += SymbolicFacetPatchBFI( form =  gamma_stab  * 
                                    mu[i] / h / h * 
                                    (u[i] - u[i].Other()) * (v[i] - v[i].Other()), 
                                    skeleton=False, definedonelements=ba_facets[i] )

    # # apply mesh adaptation 
    # if( doDeform ):   
    #     mesh.SetDeformation(deformation)

    if( kwargs['highorder'] ):
        print('setting deformation')
        mesh.SetDeformation( kwargs['deformation'] )

    # setting up matrix and vector
    a.Assemble()
    f.Assemble()

    #updating with non-homogeneous boundary conditions
    gfu = GridFunction( CutVh )
    #boundary lies in positive part
    gfu.components[1].Set( solution[1], BND )
    f.vec.data -= a.mat * gfu.vec

    return (a,f)

def IfaceUnks(CutVh, ci):
    ifdofs = BitArray(CutVh.ndof)
    ifdofs[:] = False
    hasif = BitArray( ci.GetElementsOfType(IF) )

    for el in CutVh.Elements():
        if( hasif[el.nr] ):            
            for unk in CutVh.GetDofNrs(el):
                if unk >=0:
                    ifdofs[abs(unk)] = True

    ifdofs &= CutVh.FreeDofs()    
    return ifdofs

# def IfaceUnks(CutVh, ci):
#     ifdofs = BitArray(CutVh.ndof)
#     ifdofs[:] = False
#     hasif = BitArray( ci.GetElementsOfType(IF) )

#     for el in CutVh.Elements():
#         if( hasif[el.nr] ):
#             verts = el.vertices
#             for v in verts:
#                 # print( CutVh.GetDofNrs(v) )
#                 # print( CutVh.components[0].GetDofNrs(v) )
#                 # print( CutVh.components[1].GetDofNrs(v) )
#                 # print( CutVh.components[0].ndof )
#                 # print('-------------------------------')
#                 for unk in CutVh.GetDofNrs(v):                    
#                     ifdofs[unk] = True

#     ifdofs &= CutVh.FreeDofs()
#     #print(ifdofs)
#     #input('weiter')
#     return ifdofs

class CutFemSmoother:     
    def __init__(self,**kwargs):
        self.a = kwargs["a"]
        ifdofs = IfaceUnks(kwargs["CutVh"],kwargs["ci"])
        #self.proj = Projector(mask=self.ifdofs,range=True)
        ifdofslst = [ [i] for i in range(len(ifdofs)) if ifdofs[i]==True ]
        self.proj = self.a.mat.CreateBlockSmoother(ifdofslst)
        #self.inv = a.mat.Inverse(self.ifdofs, inverse="sparsecholesky")
        self.blksm = False
        if "blocks" in kwargs:
            self.preJ = self.a.mat.CreateBlockSmoother( kwargs["blocks"] )
            self.blksm = True
        else:
            self.preJ = self.a.mat.CreateSmoother( kwargs["CutVh"].FreeDofs() )
        
        self.finestsm = kwargs.get('finest',False)
        self.loworder = kwargs.get('loworder',True)
    def __del__(self):
            print('smoother dying')
            input('destruktor')

    def IfCorr(self, u, rhs,k,nu):
        if self.loworder:
            return
        if not(self.finestsm or k==nu-1):
            #don't do interface correction
            #not last smoothing step
            return
        update = u.CreateVector()
        # this needs to be more efficient 
        # with help of ifdofs ...
        update.data = self.a.mat * u - rhs
        cgoutput = u.CreateVector()
        if self.blksm:
            cgoutput.data = CG(self.a.mat, update, pre=self.proj, tol=1e-2,printrates=False)
        else:
            cgoutput.data = CG(self.a.mat, update, pre=self.proj, tol=1e-2,printrates=False)
        #u.data -= self.inv * update
        u.data -= cgoutput

    def Smooth( self, u,rhs, nu ):
        for k in range(nu):
            self.preJ.Smooth(u, rhs)        
            self.IfCorr(u,rhs,k,nu)

    

def CutFESpace( V, ci, flags=None ):    
    hasneg = ci.GetElementsOfType(HASNEG)  
    haspos = ci.GetElementsOfType(HASPOS)

    if isinstance(V,list):
        if len(V) != 2:
            raise ValueError("you need to provide a list of two FE spaces")
        else:
            return FESpace( [   Compress(V[0],active_dofs=GetDofsOfElements(V[0],hasneg) ), 
                                Compress(V[1],active_dofs=GetDofsOfElements(V[1],haspos) ) 
                            ], flags=flags
                          ) 
    else:
        return FESpace( [   Compress(V,active_dofs=GetDofsOfElements(V,hasneg) ), 
                            Compress(V,active_dofs=GetDofsOfElements(V,haspos) ) 
                        ], flags=flags
                      ) 


class MultiGridCL(BaseMatrix):
        def __init__(self,**kwargs):
            super(MultiGridCL,self).__init__()
            self.nu     = kwargs.get('nu',2)            
            self.l      = kwargs['level']
            self.prol   = kwargs['prol']
            self.mats   = kwargs['matrices']
            self.smoothers  = kwargs['smoothers']
            self.inv    = kwargs["coarsegridsolver"]
            f = kwargs["f"]
            nref = self.l
            self.errors     = [f.vec.CreateVector() for i in range(nref+1)]
            self.defects    = [f.vec.CreateVector() for i in range(nref+1)]
        
        def __del__(self):
            print('mg dying')
            input('destruktor')


        def MultiGridIter(self,rhs,u,level):
            error   = self.errors[level]
            defect  = self.defects[level]

            if level == 0 :
                u.data = self.inv*rhs
                return
        #smoothing
            #for i in range(self.nu):
            self.smoothers[level].Smooth(u,rhs,self.nu)
        #compute defect
            defect.data = rhs - self.mats[level]*u
        #restrict defect
            self.prol.Restrict(level,defect)
            error[:] = 0
            #v-cycle
            self.MultiGridIter(defect,error,level-1)
            self.prol.Prolongate(level,error)
            u.data += error
        #smoothing
            #for i in range(self.nu):
            self.smoothers[level].Smooth(u,rhs,self.nu)
            return u

        def Mult(self,w,z):
            #z[:] = 0.0
            #z.data = w
            #z.data = 
            self.MultiGridIter(w,z,self.l)
        def Height(self):
            return a.mat.height
        def Width(self):
            return a.mat.height

def CreateLsetMeshAdapt(order):
    lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=1000, discontinuous_qn=True)
    deformation = lsetmeshadap.CalcDeformation(levelset)
    lsetp1 = lsetmeshadap.lset_p1

    return (lsetmeshadap, deformation, lsetp1 )

class LinearMGIterator:
    def __init__(self, **kwargs):
        self.nref = kwargs['nref']

        # levelset fe space
        VhL = H1(mesh, order = 1, dirichlet=[], dgjumps = True)
        lsetp1 = GridFunction(VhL)
        InterpolateToP1(levelset,lsetp1)

        ci = CutInfo(mesh, lsetp1)   
        
        # cut fe space
        Vh = H1(mesh, order=1, dirichlet=[1,2,3,4])
        VhNeg = Compress( Vh, active_dofs = GetActiveDof(mesh, HASNEG, order=1) )
        VhPos = Compress( Vh, active_dofs = GetActiveDof(mesh, HASPOS, order=1) )

        CutVh = FESpace( [VhNeg, VhPos], flags={"dgjumps": True} )

        self.CutVh = CutVh

        # construct prolongation for neg/pos space
        prolongNeg = P1Prolongation(mesh)
        prolongPos = P1Prolongation(mesh)
        # create cut prolongation
        CutProl = CompoundProlongation( CutVh )
        CutProl.AddProlongation(prolongNeg)
        CutProl.AddProlongation(prolongPos)
        prolongNeg.Update(VhNeg)
        prolongPos.Update(VhPos)
        

        #create coarse level bilinear/linear form
        self.a,self.f = AssembleCutPoisson( CutVh=CutVh, ci=ci, lsetp1=lsetp1, highorder=False )

        
        #list of coarse grid matrices
        self.mats = {}
        self.mats[0] = self.a.mat
        #list of smoothers
        self.smoothers = {}
        self.smoothers[0] = CutFemSmoother(a=self.a, CutVh=CutVh, ci=ci) 
        #coarse grid solver
        inv = self.a.mat.Inverse(CutVh.FreeDofs(), inverse="sparsecholesky")
        

        # create multilevel hierarchy
        for i in range(self.nref):
            print('preforming refinement: ', i+1)
            mesh.Refine()
            #levelset update
            VhL.Update()
            lsetp1.Update()
            InterpolateToP1(levelset,lsetp1)
            ci = CutInfo(mesh,lsetp1)
            #cut space update
            Vh.Update()
            CutVh.components[0].SetActiveDofs( GetActiveDof(mesh,HASNEG,order=1) )
            CutVh.components[1].SetActiveDofs( GetActiveDof(mesh,HASPOS,order=1) )
            CutVh.Update()
            CutProl.Update(CutVh)
            #reassemble
            self.a,self.f = AssembleCutPoisson( CutVh=CutVh, ci=ci, lsetp1=lsetp1, highorder=False )        
            self.mats[i+1] = self.a.mat

            smootherflag = (i == self.nref-1)
            print("refinement: ", i+1, ", smootherflag: ", smootherflag)
            self.smoothers[i+1] = CutFemSmoother(a=self.a, CutVh=CutVh, ci=ci, loworder=smootherflag)

        
        self.lsetp1 = lsetp1
        # create multigrid solver
        self.MGpre = MultiGridCL ( level=self.nref, prol=CutProl, 
                            matrices=self.mats, smoothers=self.smoothers,
                            coarsegridsolver=inv, f=self.f 
                            )


    def getRhs(self):
        return self.f.vec.data

    def createVec(self):
        return self.f.vec.CreateVector()

    def getSpaceDim(self):
        return [ self.CutVh.components[i].ndof for i in [0,1] ]

    def getLset(self):
        return self.lsetp1                

    def iterate(self,rhs,tol=1e-6, maxit=10,printinfo=True):
        usol = GridFunction(self.CutVh) 
        usol.vec[:] = 0.

        res = usol.vec.CreateVector()
        projres = usol.vec.CreateVector()
        
        proj = Projector(mask=self.CutVh.FreeDofs(),range=True) 
        fnew = rhs #f.vec.data     

        projres.data = proj*fnew #f.vec
        normf = Norm(projres)

        for it in range(maxit):
            usol.vec.data = self.MGpre*fnew
            res.data = fnew - self.a.mat*usol.vec
            projres.data = proj*res
            res_norm = Norm(projres) / normf
            if printinfo:
                print("it =", it+1, " ||res||_2 =", res_norm)
            if res_norm < tol:
                break        

        #udraw = GridFunction( self.CutVh )
        #udraw.components[1].Set( solution[1], BND )
        #udraw.vec.data += usol.vec

        return usol

def VertPatches(fes):
    blocks = []
    freedofs = fes.FreeDofs()
    for v in mesh.vertices:
        vdofs = set()
        for el in mesh[v].elements:
            #print( fes.GetDofNrs(el) )
            #input('w')
            vdofs |= set(d for d in fes.GetDofNrs(el) if (freedofs[abs(d)] and d >=0) )
        #print(vdofs)
        blocks.append (vdofs)
    return blocks

def ElemPatches(fes):
    blocks = []
    freedofs = fes.FreeDofs()
    for el in fes.Elements():
        eldofs = set(d for d in fes.GetDofNrs(el) if (freedofs[abs(d)] and d >=0) )        
        blocks.append (eldofs)
    return blocks

def EdgePatches(fes):
    blocks = []
    freedofs = fes.FreeDofs()
    for edge in mesh.edges:
        edofs = fes.GetDofNrs(edge)

        if( len(edofs) == 0 ): continue
        
        # print('found edge dof: ', edofs)
        elcntr = 0
        edofs = set()
        for el in mesh[edge].elements:
            elcntr += 1
            # print('  with element ', elcntr, ' el: ', fes.GetDofNrs(el) )
            # print('  with element ', elcntr, ' el: ', el.nr )
            edofs |= set( d for d in fes.GetDofNrs(el) if(freedofs[abs(d)] and d>=0) )
    
        # input('next edge')
        blocks.append (edofs)
    return blocks

class HoTwoGridCL:
    def __init__(self,**kwargs):
        self.a = kwargs['a']
        self.f = kwargs['f']
        self.fes = kwargs['fes']
        # blocks = ElemPatches(self.fes)
        blocks = EdgePatches(self.fes)
        # blocks = VertPatches(self.fes)
        self.blockjac = CutFemSmoother( a=self.a, CutVh=self.fes, 
                                        ci=kwargs["ci"], blocks=blocks, 
                                        finest=True, loworder=False )
        #self.a.mat.CreateBlockSmoother(blocks)
        #self.smoother = GaussSeid(self.blockjac)
        self.mgiter = kwargs['mgiter']

    def iterate(self,nu=2,maxit=20,tol=1e-6):
        vertexdofs = BitArray(self.fes.ndof)
        vertexdofs[:] = False

        for v in mesh.vertices:
            #print( self.fes.GetDofNrs(v) )
            for d in self.fes.GetDofNrs(v):
                if d>=0:
                    vertexdofs[d] = True
                
        vertexdofs &= self.fes.FreeDofs()

        coarseProj = Projector( mask= vertexdofs, range=True)
        
        proj = Projector(mask=self.fes.FreeDofs(),range=True)
        usol = GridFunction(self.fes)
        usol.vec[:] = 0.

        projres = usol.vec.CreateVector()
        projres.data = proj*self.f.vec
        normf = Norm(projres)

        LinDof = self.mgiter.getSpaceDim()
        HOdof  = [ self.fes.components[i].ndof for i in [0,1] ]

        oldres = Norm(projres)

        for it in range(maxit):
            # initial smoothing 
            #for sm in range(nu):
            self.blockjac.Smooth(usol.vec, self.f.vec,nu)

            res = usol.vec.CreateVector()
            # perform multiplication more efficiently ...
            # only 'upper' part (vertex dofs) needed 
            res.data = self.f.vec -self.a.mat * usol.vec
            
            projres.data = coarseProj*res # i don't think this is really needed... data copied anyways ...

            # print(projres.Range(0, LinDof[0] ))
            # input('weiter')
            # print(projres.Range(HOdof[0], HOdof[0] + LinDof[1] ))
            # input('weiter')

            mgrhs = self.mgiter.createVec()
            # copy data from two domains 
            # ... should be done in combination with vertexdofs ...
            mgrhs.Range(0, LinDof[0]).data = projres.Range(0, LinDof[0] )
            mgrhs.Range( LinDof[0], len(mgrhs) ).data = projres.Range(HOdof[0], HOdof[0] + LinDof[1] )
            #print('lens: ', len(mgrhs), LinDof[0], LinDof[1], LinDof[0]+LinDof[1])

            # coarse grid correction: multigrid on linears
            cup = self.mgiter.iterate(mgrhs,maxit=1,printinfo=False)

            # update solution
            # ... should be done in combination with vertexdofs ...
            usol.vec.Range(0, LinDof[0] ).data += cup.vec.Range(0, LinDof[0])
            usol.vec.Range( HOdof[0], HOdof[0] + LinDof[1]).data += cup.vec.Range( LinDof[0], len(mgrhs) )

            # compute residual for measurement
            res.data = self.f.vec - self.a.mat*usol.vec
            projres.data = proj*res
            res_norm = Norm(projres)
            
            print("tg-it =", it+1, "\t ||res||_2 = {0:.2E}".format(res_norm), "\t reduction: {0:.2f}".format(res_norm/oldres))
            if res_norm < tol*normf:
                break
            oldres = res_norm

        udraw = GridFunction( self.fes )
        udraw.components[1].Set( solution[1], BND )
        udraw.vec.data += usol.vec
        return udraw



if __name__ == '__main__':
    main()   



"""
	Description:
	------------
	3D version of zalesak_disk.py
"""

from netgen.csg import *
from ngsolve import *
from xfem import *


geo = CSGeometry()
geo.Add(OrthoBrick(Pnt(-1,-1,-1), Pnt(1,1,1)))
mesh = Mesh(geo.GenerateMesh(maxh=0.2))


V = H1(mesh,order=1)
lsetp1_1, lsetp1_2, lsetp1_3 = GridFunction(V), GridFunction(V), GridFunction(V)

InterpolateToP1( x**2 + y**2 + z**2 - 0.8**2, lsetp1_1)
InterpolateToP1( 0.4 - y, lsetp1_2)
InterpolateToP1( 0.2 - IfPos(x,x,-x), lsetp1_3)


Draw(lsetp1_1,mesh,"lsetp1_1")
Draw(lsetp1_2,mesh,"lsetp1_2")
Draw(lsetp1_3,mesh,"lsetp1_3")

inner = {"levelsets": (lsetp1_1, lsetp1_2, lsetp1_3),
		 "domain_type": (NEG,ANY,NEG)|(NEG,NEG,ANY)
		}

boundary = {"levelsets": (lsetp1_1, lsetp1_2, lsetp1_3),
			"domain_type": (NEG,POS,IF)|(NEG,IF,NEG)|(IF,ANY,NEG)|(IF,NEG,ANY)
		   }


import pytest
from ngsolve.meshes import *
from ngsolve import *
from xfem import *
from math import pi
from xfem.lsetcurv import *

@pytest.mark.parametrize("domain", [NEG, POS, IF])
@pytest.mark.parametrize("alpha", [2,4,8])

def test_polynomial_ET_Segm(domain, alpha):
    order = alpha
    mesh = Make1DMesh(1)
    
    x_ast = 0.78522
    levelset = x_ast-x
    referencevals = {POS: x_ast**(alpha+1)/(alpha+1), NEG:(1-x_ast**(alpha+1))/(alpha+1), IF: x_ast**(alpha)}
    V = H1(mesh, order=1)
    lset_approx = GridFunction(V)
    #InterpolateToP1(levelset,lset_approx)
    lset_approx.Set(levelset)
    
    f = x**alpha
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    print("Result of Integration Key ",domain," : ", integral)
    error = abs(integral - referencevals[domain])
    print("Error: ", error)
    
    assert error < 5e-15*(order+1)*(order+1)

@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("order", [2,4,8])
@pytest.mark.parametrize("domain", [NEG, POS, IF])

def test_new_integrateX_via_circle_geom(quad, order, domain):
    r=0.6

    levelset = sqrt(x*x+y*y)-r
    referencevals = { POS : 1-pi*r*r/4, NEG : pi*r*r/4, IF : r*pi/2}

    n_ref = 8
    errors = []

    for i in range(n_ref):
        mesh = MakeStructured2DMesh(quads = quad, nx=2**i, ny=2**i)    

        V = H1(mesh,order=1)
        lset_approx = GridFunction(V)
        InterpolateToP1(levelset,lset_approx)
    
        f = CoefficientFunction(1)
    
        integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                             cf=f, mesh=mesh, order = order)
        print("Result of Integration Reflevel ",i,", Key ",domain," : ", integral)
        errors.append(abs(integral - referencevals[domain]))

        
    eoc = [log(errors[i+1]/errors[i])/log(0.5) for i in range(n_ref-1)]

    print("L2-errors:", errors)
    print("experimental order of convergence (L2):", eoc)

    mean_eoc_array = eoc[1:]
    mean_eoc = sum(mean_eoc_array)/len(mean_eoc_array)
    assert mean_eoc > 1.75

@pytest.mark.parametrize("order", [2,4])
@pytest.mark.parametrize("domain", [POS, NEG])
def test_new_integrateX_via_sphere_geom_quad(order, domain):
    r=0.7234436998

    levelset = sqrt(x*x+y*y+z*z)-r
    referencevals = { POS : 1-pi*r*r*r/6, NEG : pi*r*r*r/6, IF : r*r*pi/2}

    n_ref = 6
    errors = []

    for i in range(n_ref):
        mesh = MakeStructured3DMesh(hexes = True, nx=2**i, ny=2**i, nz=2**i)
        print("i: " +str(i))
        print("Argument Meshing: ",str(2**i))
        
        V = H1(mesh,order=1)
        lset_approx = GridFunction(V)
        InterpolateToP1(levelset,lset_approx)
    
        f = CoefficientFunction(1)
    
        integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
        print("Result of Integration Reflevel ",i,", Key ",domain," : ", integral)
        errors.append(abs(integral - referencevals[domain]))
        
    eoc = [log(errors[i+1]/errors[i])/log(0.5) for i in range(n_ref-1)]

    print("L2-errors:", errors)
    print("experimental order of convergence (L2):", eoc)

    mean_eoc_array = eoc[1:]
    mean_eoc = sum(mean_eoc_array)/len(mean_eoc_array)
    assert mean_eoc > 1.75
    
@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("order", [2,4,8])
@pytest.mark.parametrize("domain", [NEG, POS, IF])
@pytest.mark.parametrize("N", [1,10,30])

def test_new_integrateX_via_straight_cutted_quad2D(order, domain, quad, N):
    mesh = MakeStructured2DMesh(quads = quad, nx=N, ny=N)    
    
    levelset = 1 - 2*x - 2*y
    referencevals = {NEG: 7/8, POS: 1/8, IF: 1/sqrt(2)}
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)

@pytest.mark.parametrize("quad", [False, True])
@pytest.mark.parametrize("order", [2])
@pytest.mark.parametrize("domain", [IF, NEG, POS])
@pytest.mark.parametrize("dim", [x,y])
@pytest.mark.parametrize("eps", [1e-1, 1e-2, 5e-3, 1e-3, 0])

def test_new_integrateX_via_orth_cutted_quad2D_epsiloned(order, domain, quad, dim, eps):
    mesh = MakeStructured2DMesh(quads = quad, nx=1, ny=1)    
    
    if dim == x:
        levelset = 1 - 2*x + eps*(y-0.5)
    elif dim == y:
        levelset = 1 - 2*y + eps*(x-0.5)
        
    referencevals = {NEG: 1./2, POS: 1./2, IF: sqrt(1.+eps*eps/4) }
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    error = abs(integral - referencevals[domain])
    print(error)
    
    assert error < 5e-15*(order+1)*(order+1)


@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("order", [2,4,8])
@pytest.mark.parametrize("domain", [NEG, POS])
@pytest.mark.parametrize("alpha", [0,1,2])
@pytest.mark.parametrize("dim", [x,y])

#integrate f(x) = dim^alpha on the geometry implied by phi(x,y,z) = 1 - 2*x - 2*y
# for analytic solution see
# http://www.wolframalpha.com/input/?i=integrate+from+0+to+1%2F2+from+0+to+(1%2F2-x)+x%5Ealpha+dy+dx
def test_new_integrateX_via_straight_cutted_quad2D_polynomial(order, domain, quad, alpha, dim):
    mesh = MakeStructured2DMesh(quads = quad, nx=1, ny=1)    
    # square = SplineGeometry()
    # square.AddRectangle([0,0],[1,1],bc=1)
    # mesh = Mesh (square.GenerateMesh(maxh=100, quad_dominated=quad))
    
    levelset = 1 - 2*x - 2*y
    val_pos = 2**(-alpha-2)/(alpha*alpha + 3*alpha+2)
    referencevals = {POS: val_pos, NEG: 1./(alpha+1) - val_pos}
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = dim**alpha
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)

@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("order", [2,4,8])
@pytest.mark.parametrize("domain", [NEG, POS, IF])

def test_new_integrateX_via_straight_cutted_quad3D(order, domain, quad):
    mesh = MakeStructured3DMesh(hexes = quad, nx=1, ny=1, nz=1)    
    
    levelset = 1 - 2*x - 2*y - 2*z
    referencevals = { POS : 1./48, NEG : 47./48, IF : sqrt(3)/8 }
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    print("Integral: ", integral)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)

@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("order", [4])
@pytest.mark.parametrize("domain", [NEG, POS])
@pytest.mark.parametrize("alpha", [0,1,2])
@pytest.mark.parametrize("dim", [x,y,z])

#integrate f(x) = dim^alpha on the geometry implied by phi(x,y,z) = 1 - 2*x - 2*y - 2*z
# for analytic solution see
# http://www.wolframalpha.com/input/?i=integrate+from+0+to+1%2F2+from+0+to+(1%2F2-x)+from+0+to+(1%2F2-x-y)+x%5Ealpha+dz+dy+dx
def test_new_integrateX_via_straight_cutted_quad3D_polynomial(order, domain, quad, alpha, dim):
    ngsglobals.msg_level = 0
    mesh = MakeStructured3DMesh(hexes = quad, nx=1, ny=1, nz=1)    
        
    levelset = 1 - 2*x- 2*y - 2*z
    val_pos = 2**(-alpha-3)/(alpha**3+6*alpha*alpha + 11*alpha+6)
    referencevals = {POS: val_pos, NEG: 1./(alpha+1) - val_pos}
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = dim**alpha
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)

@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("order", [2])
@pytest.mark.parametrize("domain", [NEG, POS, IF])
@pytest.mark.parametrize("dim", [x,y])

def test_new_integrateX_via_orth_cutted_quad2D(order, domain, quad, dim):
    mesh = MakeStructured2DMesh(quads = quad, nx=1, ny=1)    
    
    levelset = 1 - 3*dim
    referencevals = {NEG: 2./3, POS: 1./3, IF: 1. }
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)
    

@pytest.mark.parametrize("quad", [False, True])
@pytest.mark.parametrize("order", [2])
@pytest.mark.parametrize("domain", [NEG, POS])
@pytest.mark.parametrize("dim", [x,y,z])

def test_new_integrateX_via_orth_cutted_quad3D(order, domain, quad, dim):
    mesh = MakeStructured3DMesh(hexes = quad, nx=1, ny=1, nz=1)    
    
    levelset = 1 - 2*dim
    referencevals = { POS : 1./2, NEG : 1./2, IF : 1. }
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)

def test_eb_cut_integrator_2d():
    from netgen.geom2d import SplineGeometry

    ngsglobals.msg_level = 1

    levelset = (x -1) **4 + (y -1)**4 + 12*y + 1.4*(x -2)**3 - 15
    len_box = 4

    exact = sin(y)
    exact_grad = CoefficientFunction((0, cos(y)))

    order = 5

    l2errors = []
    h1errors = []

    extend_Fh = False
    condense = True

    for i in [2,3,4]:
        square = SplineGeometry()
        square.AddRectangle([-len_box,-len_box],[len_box,len_box],bc=1)
        mesh = Mesh (square.GenerateMesh(maxh=len_box/1.5*0.5**(i), quad_dominated=False))
        print("Max_h = ", len_box/1.5*0.5**(i))

        lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=10.25, discontinuous_qn=True)
        deformation = lsetmeshadap.CalcDeformation(levelset)
        lsetp1 = lsetmeshadap.lset_p1
            
        # TraceFESpace 
        fes1 = L2(mesh, order=order)
        fes2 = FacetFESpace(mesh, order=order)
        fes3 = FacetFESpace(mesh, order=order - 1)
        
        Vhg = FESpace([fes1, fes2, fes3], dgjumps=not condense)
        
        # overwrite freedofs of VhG to mark only dofs that are involved in the cut problem
        ci = CutInfo(mesh, lsetp1)
        reg_Th = ci.GetElementsOfType(IF)   
        reg_Fh = GetFacetsWithNeighborTypes(mesh,a=reg_Th,b=reg_Th,use_and=not extend_Fh)

        gf_reg_Fh = GridFunction(FacetFESpace(mesh,order=0))
        for i in range(len(reg_Fh)):
            gf_reg_Fh.vec[i] = 1 if reg_Fh[i] else 0
        
        if not extend_Fh:
                if not condense:
                    freedofs = CompoundBitArray( [GetDofsOfElements(fes1,reg_Th),  GetDofsOfFacets(fes2,reg_Fh) ,  GetDofsOfFacets(fes3,reg_Fh) ] )
                else:
                    freedofs = CompoundBitArray( [fes1.FreeDofs(True) & GetDofsOfElements(fes1,reg_Th),  GetDofsOfFacets(fes2,reg_Fh) ,  GetDofsOfFacets(fes3,reg_Fh) ] )
        else:
                if not condense:
                    freedofs = GetDofsOfElements(Vhg,reg_Th)
                else:
                    freedofs = CompoundBitArray( [fes1.FreeDofs(True) & GetDofsOfElements(fes1,reg_Th),  GetDofsOfElements(fes2,reg_Th) ,  GetDofsOfElements(fes3,reg_Th) ] )
        
        normal_helper_gf = GridFunction(HDiv(mesh, order=0))
        for i in range(len(normal_helper_gf.vec)):
            normal_helper_gf.vec[i] = 1
        
        gfu = GridFunction(Vhg)
            
        #tangential projection to given normal
        def P(u,n_phi):
            return u - (u*n_phi)*n_phi
        
        #normalization (pointwise) of a vector
        def Normalized(u):
            return 1.0 / Norm(u) * u
        
        n_phi1 = Normalized(grad(lsetp1))
        n_phi2 = Normalized(grad(lsetp1).Other())
        
        h = specialcf.mesh_size
        n_F = specialcf.normal(2)
        
        conormal1 = Normalized(P(n_F,n_phi1))
        conormal2 = Normalized(P(-n_F,n_phi2))
        
        normal_helper_proj = InnerProduct(normal_helper_gf, n_F)*n_F
        normal_helper = Normalized(normal_helper_proj)
            
        def jump(u, uhat):
            return u - uhat
        
        beta_E = 4 * (order+1)**2
        beta_F = 100.
        beta_F2 = 1.
        lam_nd = 0.0 if order == 1 else 0.1*1./h + 0.1*h
        
        u, uhat, sigmahat = Vhg.TrialFunction()
        v, vhat, tauhat = Vhg.TestFunction()
        lset_if  = { "levelset" : lsetp1, "domain_type" : IF , "subdivlvl" : 0}
        
        a = RestrictedBilinearForm(Vhg,"a",reg_Th,reg_Fh,check_unused=False, flags={"eliminate_internal": condense})
        a += SymbolicBFI(levelset_domain = lset_if, form = P(grad(u),n_phi1) * P(grad(v),n_phi1) + u * v, definedonelements=reg_Th)
        a += SymbolicBFI(form = (lam_nd * grad(u)*n_phi1) * (grad(v)*n_phi1), definedonelements=reg_Th)
        
        a += SymbolicBFI(levelset_domain = lset_if, form = ( - InnerProduct(grad(u),conormal1)*jump(v, vhat)
                                                                - InnerProduct(grad(v),conormal1)*jump(u, uhat)
                                                                + beta_E/h * (u- uhat)*(v - vhat) ) *gf_reg_Fh,
                                                                element_boundary=True, definedonelements=reg_Th)
        
        a += SymbolicBFI(form = beta_F/(h*h) * (u- uhat) * (v - vhat) *gf_reg_Fh, element_boundary=True, definedonelements=reg_Th)

        a += SymbolicBFI(form = beta_F2 * (grad(u)*normal_helper- sigmahat) * (grad(v)*normal_helper - tauhat) *gf_reg_Fh, element_boundary=True, definedonelements=reg_Th)

        f_coeff = -(4*(y - 1)**3 + 12)**2*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-1.0)*sin(y) + (4*(y - 1)**3 + 12)*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-0.5)*(-12.0*(y - 1)**2*(4*(y - 1)**3 + 12)**2*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-1.5) + 12*(y - 1)**2*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-0.5) - 0.5*(4.2*(x - 2)**2 + 4*(x - 1)**3)**2*((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-1.5)*(16.8*x + 24*(x - 1)**2 - 33.6) + ((4.2*(x - 2)**2 + 4*(x - 1)**3)**2 + (4*(y - 1)**3 + 12)**2)**(-0.5)*(8.4*x + 12*(x - 1)**2 - 16.8))*cos(y) + 2*sin(y)
        
        f = LinearForm(Vhg)
        f += SymbolicLFI(levelset_domain = lset_if, form = f_coeff * v, definedonelements=reg_Th)

        mesh.SetDeformation(deformation)
        a.Assemble()
        f.Assemble();
        
        gfu.vec[:] = 0.0
        if not condense:
            gfu.vec.data = a.mat.Inverse(freedofs, "sparsecholesky") * f.vec
        else:
            f.vec.data += a.harmonic_extension_trans * f.vec 
            inv = a.mat.Inverse(freedofs, "sparsecholesky")
            gfu.vec.data = inv * f.vec
            
            gfu.vec.data += a.harmonic_extension * gfu.vec
            gfu.vec.data += a.inner_solve * f.vec
        
        err_sqr_coefs = (gfu.components[0]-exact)**2
        l2error = sqrt( Integrate( levelset_domain=lset_if, cf=err_sqr_coefs, mesh=mesh, order=2*order+1) )
        print ("l2error : ", l2error)
        l2errors.append(l2error)
        
        H1_err_sqr_coefs = Norm( P (gfu.components[0].Deriv() - exact_grad, n_phi1) )**2
        h1error = sqrt( l2error**2 + Integrate( levelset_domain=lset_if, cf=H1_err_sqr_coefs, mesh=mesh, order=2*order+1) )
        print("h1error : ", h1error)
        h1errors.append(h1error)
        mesh.UnsetDeformation()
    
    assert l2error < 1e-7
    assert h1error < 1e-5

    if len(l2errors) > 1:
        eocs = [log(l2errors[i-1]/l2errors[i])/log(2) for i in range(1,len(l2errors))]
        print ("l2 eocs : ", eocs)
        assert sum(eocs)/len(eocs) > 5.5
        
        eocs = [log(h1errors[i-1]/h1errors[i])/log(2) for i in range(1,len(h1errors))]
        print ("h1 eocs : ", eocs)
        assert sum(eocs)/len(eocs) > 4.5



"""
In this example we solve a scalar *unfitted* interface problem as
in `demos/unf_interf_prob.py`, but in an MPI-compatible version.

Domain + PDE problem + Discretization:
--------------------------------------
As `demos/unf_interf_prob.py` in (but only the XFEM version)

Implementational aspects:
-------------------------
* Geometry approximation: As in `demos/unf_interf_prob.py`

* Linear systems: linear systems are solved with a CG solver and an algebraic
  multigrid preconditioner.

* The solution is written to VTK files which allows visualization with
  e.g. paraview.

"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from math import pi
import os
from mpi4py import MPI

import netgen.meshing
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem.lsetcurv import *
from xfem import *
comm = MPI.COMM_WORLD
rank = comm.rank
np = comm.size
if np > 1:
    from ngsolve import ngs2petsc

ngsglobals.msg_level = 10

# -------------------------------- PARAMETERS ---------------------------------
# Domain corners
ll, ur = (-1.5, -1.5), (1.5, 1.5)
# Mesh diameter
maxh = 0.2
# Finite element space polynomial order
order = 2
# diffusion coefficients for the sub-domains (NEG/POS):
alpha = [1.0, 2.0]
# Nitsche penalty parameter
lambda_nitsche = 20
# Write VTK output for visualisation
do_vtk = True


# ----------------------------------- MAIN ------------------------------------
# We generate the background mesh of the domain and use a simplicity
# triangulation to obtain a mesh with quadrilaterals use
# 'quad_dominated=True'
geo = SplineGeometry()
geo.AddRectangle(ll, ur, bc=1)

if rank == 0:
    ngmesh = geo.GenerateMesh(maxh=maxh)
    if np > 1:
        ngmesh.Distribute(comm)
else:
    ngmesh = netgen.meshing.Mesh.Receive(comm)
    ngmesh.SetGeometry(geo)

mesh = Mesh(ngmesh)

# Level set function of the domain and higher-order machinery
levelset = (sqrt(sqrt(x**4 + y**4)) - 1)

lsetadap = LevelSetMeshAdaptation(mesh, order=order, threshold=1000,
                                  discontinuous_qn=True)
deformation = lsetadap.CalcDeformation(levelset)
lsetp1 = lsetadap.lset_p1
Draw(lsetp1, mesh, "lsetp1")

mesh.SetDeformation(deformation)

# Gather information of cut elements
ci = CutInfo(mesh, lsetp1)

# Extended finite element space
Vh = H1(mesh, order=order, dirichlet=[1, 2, 3, 4])
Vhx = XFESpace(Vh, ci)

VhG = FESpace([Vh, Vhx])
print("unknowns in extended FESpace for rank", rank, ":", VhG.ndof)

# Manufactured solution and corresponding r.h.s. data CoefficientFunctions:
r22 = x**2 + y**2
r44 = x**4 + y**4
r66 = x**6 + y**6
r41 = sqrt(sqrt(r44))
r4m3 = 1.0 / r41**3
solution = [1 + pi / 2 - sqrt(2.0) * cos(pi / 4 * r44), pi / 2 * r41]
coef_f = [-alpha[i] * (solution[i].Diff(x).Diff(x)
                       + solution[i].Diff(y).Diff(y)) for i in range(2)]

# Coefficients:
n = 1.0 / grad(lsetp1).Norm() * grad(lsetp1)
h = specialcf.mesh_size

# The cut ratio extracted from the cutinfo-class
kappa = (CutRatioGF(ci), 1.0 - CutRatioGF(ci))
# Nitsche stabilization parameter:
stab = lambda_nitsche * (alpha[1] + alpha[0]) / h

# Expressions of test and trial functions:
u_std, u_x = VhG.TrialFunction()
v_std, v_x = VhG.TestFunction()

u = [u_std + op(u_x) for op in [neg, pos]]
v = [v_std + op(v_x) for op in [neg, pos]]

gradu = [grad(u_std) + op(u_x) for op in [neg_grad, pos_grad]]
gradv = [grad(v_std) + op(v_x) for op in [neg_grad, pos_grad]]

average_flux_u = sum([- kappa[i] * alpha[i] * gradu[i] * n for i in [0, 1]])
average_flux_v = sum([- kappa[i] * alpha[i] * gradv[i] * n for i in [0, 1]])

# Integration domains for integration on negative/positive sub-domains
# and on the interface: Here, the integration is (geometrically) exact
# if the "levelset"-argument is a piecewise (multi-)linear function.
# We further provide a mesh deformation that is applied in the higher order
# case:
dx = tuple([dCut(lsetp1, dt, deformation=lsetadap.deform,
                 definedonelements=ci.GetElementsOfType(HAS(dt)))
            for dt in [NEG, POS]])
ds = dCut(lsetp1, IF, deformation=lsetadap.deform)

# bilinear form for the unfitted Nitsche formulation:
a = BilinearForm(VhG, symmetric=True)
a += sum(alpha[i] * gradu[i] * gradv[i] * dx[i] for i in [0, 1])
a += (average_flux_u * (v[0] - v[1]) + average_flux_v * (u[0] - u[1])
      + stab * (u[0] - u[1]) * (v[0] - v[1])) * ds

# R.h.s.:
f = LinearForm(VhG)
f += sum(coef_f[i] * v[i] * dx[i] for i in [0, 1])

# Solution vector
gfu = GridFunction(VhG)
gfu.components[0].Set(solution[1], BND)

# # setting up matrix and vector
# c = Preconditioner(a, 'hypre') # very good for low order but fine meshes
# c = Preconditioner(a, 'bddc') # very good for high order but moderate meshes
# c = Preconditioner(a, 'direct', inverse="masterinverse") # only for small run
# c = Preconditioner(a, 'gamg') # petsc gamg
if np == 1:
    c = Preconditioner(a, 'bddc', coarsetype="h1amg")  # only for serial runs
else:
    c = Preconditioner(a, 'bddc', coarsetype="gamg")

a.Assemble()
f.Assemble()

rhs = gfu.vec.CreateVector()
rhs.data = f.vec - a.mat * gfu.vec
update = gfu.vec.CreateVector()
update.data = solvers.CG(mat=a.mat, pre=c.mat, rhs=rhs,
                         tol=1e-6, maxsteps=100, printrates=comm.rank == 0)
gfu.vec.data += update

uh = [gfu.components[0] + op(gfu.components[1]) for op in [neg, pos]]

err_sqr = sum([(uh[i] - solution[i])**2 * dx[i].order(2 * order)
               for i in [0, 1]])
l2error = sqrt(Integrate(err_sqr, mesh))

if rank == 0:
    print("L2 error : ", l2error)

# Do VTK output for visualisation
output_path = os.path.dirname(os.path.realpath(__file__)) + "/output"
if rank == 0 and not os.path.exists(output_path):
    os.mkdir(output_path)
comm.Barrier()  # wait until master has created the directory!!

u = [gfu.components[0] + op(gfu.components[1]) for op in [neg, pos]]

if do_vtk:
    vtk_output = [u[0], u[1], lsetp1, deformation]
    vtk_names = ["u1", "u2", "lset", "deformation"]

    vtk = VTKOutput(ma=mesh, coefs=vtk_output, names=vtk_names,
                    filename=output_path + "/vtkout_p" + str(rank) + "_n0",
                    subdivision=1)
    vtk.Do()
comm.Barrier()



import pytest
from ngsolve import *
from xfem import *
from netgen.geom2d import SplineGeometry
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from math import pi

@pytest.mark.parametrize("quad", [False])
@pytest.mark.parametrize("order", [2,8])
@pytest.mark.parametrize("domain", [NEG, POS, IF])

def test_new_integrateX_via_circle_geom(quad, order, domain):
    square = SplineGeometry()
    square.AddRectangle([0,0],[1,1],bc=1)
    mesh = Mesh (square.GenerateMesh(maxh=100, quad_dominated=quad))
    r=0.6

    levelset = sqrt(x*x+y*y)-r
    referencevals = { POS : 1-pi*r*r/4, NEG : pi*r*r/4, IF : r*pi/2}

    n_ref = 8
    errors = []

    for i in range(n_ref):
        f = CoefficientFunction(1)
    
        integral = Integrate(levelset_domain = { "levelset" : levelset, "domain_type" : domain},
                             cf=f, mesh=mesh, order = order)
        print("Result of Integration Reflevel ",i,", Key ",domain," : ", integral)
        errors.append(abs(integral - referencevals[domain]))

        if i < n_ref - 1:
            mesh.Refine()
        
    eoc = [log(errors[i+1]/errors[i])/log(0.5) for i in range(n_ref-1)]

    print("L2-errors:", errors)
    print("experimental order of convergence (L2):", eoc)

    mean_eoc_array = eoc[1:]
    mean_eoc = sum(mean_eoc_array)/len(mean_eoc_array)
    assert mean_eoc > 1.75

@pytest.mark.parametrize("quad", [False])
@pytest.mark.parametrize("order", [2,8])
@pytest.mark.parametrize("domain", [NEG, POS, IF])

def test_new_integrateX_via_straight_cutted_quad2D(order, domain, quad):
    square = SplineGeometry()
    square.AddRectangle([0,0],[1,1],bc=1)
    mesh = Mesh (square.GenerateMesh(maxh=100, quad_dominated=quad))
    
    levelset = 1 - 2*x - 2*y
    referencevals = {NEG: 7/8, POS: 1/8, IF: 1/sqrt(2)}
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : levelset, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)

@pytest.mark.parametrize("quad", [False])
@pytest.mark.parametrize("order", [2,8])
@pytest.mark.parametrize("domain", [NEG, POS, IF])

def test_new_integrateX_via_straight_cutted_quad3D(order, domain, quad):
    cube = OrthoBrick( Pnt(0,0,0), Pnt(1,1,1) ).bc(1)
    geom = CSGeometry()
    geom.Add (cube)
    ngmesh = geom.GenerateMesh(maxh=1.3, quad_dominated=quad)
    mesh = Mesh(ngmesh)
    
    levelset = 1 - 2*x - 2*y - 2*z
    referencevals = { POS : 1./48, NEG : 47./48, IF : sqrt(3)/8 }
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : levelset, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)



import pytest
from ngsolve import *
from ngsolve.meshes import *
from xfem import *
from xfem.lsetcurv import *
from math import pi

@pytest.mark.parametrize("quad", [True])
@pytest.mark.parametrize("order", [2,4,8])
@pytest.mark.parametrize("domain", [NEG, POS, IF])

def test_new_integrateX_via_straight_cutted_quad3D(order, domain, quad):
    mesh = MakeStructured3DMesh(hexes = True, nx=2, ny=2, nz=2)    
    
    levelset = 1 - 2*x - 2*y - 2*z
    referencevals = { POS : 1./48, NEG : 47./48, IF : sqrt(3)/8 }
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    print("Integral: ", integral)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)


@pytest.mark.parametrize("quad", [True])
@pytest.mark.parametrize("order", [4])
@pytest.mark.parametrize("domain", [NEG, POS])
@pytest.mark.parametrize("alpha", [0,1,2])
@pytest.mark.parametrize("dim", [x,y,z])

#integrate f(x) = dim^alpha on the geometry implied by phi(x,y,z) = 1 - 2*x - 2*y - 2*z
# for analytic solution see
# http://www.wolframalpha.com/input/?i=integrate+from+0+to+1%2F2+from+0+to+(1%2F2-x)+from+0+to+(1%2F2-x-y)+x%5Ealpha+dz+dy+dx
def test_new_integrateX_via_straight_cutted_quad3D_polynomial(order, domain, quad, alpha, dim):
    #ngsglobals.msg_level = 0
    mesh = MakeStructured3DMesh(hexes = quad, nx=5, ny=5, nz = 5)    

    levelset = 1 - 2*x- 2*y - 2*z
    val_pos = 2**(-alpha-3)/(alpha**3+6*alpha*alpha + 11*alpha+6)
    referencevals = {POS: val_pos, NEG: 1./(alpha+1) - val_pos}
    print("Val_pos: ", val_pos)
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = dim**alpha
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain, "quad_dir_policy" : OPTIMAL},
                         cf=f, mesh=mesh, order = order)
    print("Integral:", integral)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)

def test_new_integrateX_via_straight_cutted_quad3D_polynomial_zero_val_challenge(order=4, domain=POS, alpha=2, dim=x):
    ngsglobals.msg_level = 0

    mesh = MakeStructured3DMesh(hexes = True, nx=2, ny=2, nz= 2)    
        
    levelset = 1 - 2*x- 2*y - 2*z
    val_pos = 2**(-alpha-3)/(alpha**3+6*alpha*alpha + 11*alpha+6)
    referencevals = {POS: val_pos, NEG: 1./(alpha+1) - val_pos}
    print("Val_pos: ", val_pos)
    
    lset_approx = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lset_approx)
    
    f = dim**alpha
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain, "quad_dir_policy" : OPTIMAL},
                         cf=f, mesh=mesh, order = order)
    print("Integral:", integral)
    error = abs(integral - referencevals[domain])
    
    assert error < 5e-15*(order+1)*(order+1)
    
@pytest.mark.parametrize("quad", [True])
@pytest.mark.parametrize("order", [2,4,6])
@pytest.mark.parametrize("domain", [NEG, POS, IF])

def test_new_integrateX_TPMC_case_quad3D(order, domain, quad):
    mesh = MakeStructured3DMesh(hexes = True, nx=1, ny=1, nz = 1)    
    lset_approx = GridFunction(H1(mesh,order=1))
    for i,v in enumerate([-4,4,-1,-1,2,-3,5,-1]):
        lset_approx.vec[i] = v
    
    f = CoefficientFunction(1)
    
    integral = Integrate(levelset_domain = { "levelset" : lset_approx, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
    print("Integral: ", integral)

    if domain == IF:
        assert integral < 10
    elif domain == NEG:
        assert abs(integral - 0.5167820912197415) < 0.75
    else:
        assert abs(integral - 0.4825797907263282) < 0.75

@pytest.mark.parametrize("quad", [True])
@pytest.mark.parametrize("order", [2,4])
@pytest.mark.parametrize("high_order", [False, True])

def test_new_integrateX_TPMC_case_quad3D2(order, quad, high_order):

    mesh = MakeStructured3DMesh(hexes = quad, nx=10, ny=10, nz = 10)    
    
    #phi = -4*(1-x)*(1-y)*(1-z) + 4*(1-x)*(1-y)*z -1*(1-x)*y*(1-z) - 1*(1-x)*y*z + 2*x*(1-y)*(1-z) -3 *x*(1-y)*z + 5 * x * y * (1-z) -1 *x *y*z
    phi = x *((7*y - 13) *z + 6) + y *(3 - 8 *z) + 8 *z - 4
    
    if high_order:
        print("Creating LevelSetMeshAdaptation class")
        lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.2, discontinuous_qn=True)
        lsetp1 = lsetmeshadap.lset_p1
        deformation = lsetmeshadap.CalcDeformation(phi)
    
        mesh.SetDeformation(deformation)    
    else:
        lsetp1 = GridFunction(H1(mesh,order=1))
        lsetp1.Set(phi)
    
    f = CoefficientFunction(1)
    
    print("Doing integration")
    for domain in [POS, NEG, IF]:
        integral = Integrate(levelset_domain = { "levelset" : lsetp1, "domain_type" : domain},
                         cf=f, mesh=mesh, order = order)
        print("Integral: ", integral, " ; domain = ", domain)

        if domain == IF:
            assert abs( integral - 1.82169) < 5e-3
        elif domain == NEG:
            assert abs(integral - 0.51681) < 1e-3
        else:
            assert abs(integral - 0.48319) < 1e-3



"""
Test DomainTypeArray functionality
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
import pytest
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

ngsglobals.msg_level = 1


# -----------------------------------------------------------------------------
# ------------------------------ TEST TRIANGLE --------------------------------
# -----------------------------------------------------------------------------
def test_trianlge():
    nr_ls = 3
    triangle = DomainTypeArray([(NEG, NEG, NEG)])

    invert_traingle = ~triangle
    target = DomainTypeArray([(NEG, NEG, POS), (NEG, POS, NEG), (NEG, POS, POS),
                              (POS, NEG, NEG), (POS, NEG, POS), (POS, POS, NEG), 
                              (POS, POS, POS)])
    assert invert_traingle == target

    triangle_bnd = triangle.Boundary()
    target = DomainTypeArray([(IF, NEG, NEG), (NEG, IF, NEG), (NEG, NEG, IF)])
    assert triangle_bnd == target
    assert triangle_bnd.codim == 1

    target = target | triangle_bnd
    assert target == triangle_bnd

    target = target & triangle_bnd
    assert target == triangle_bnd

    triangle_bnd_bnd = triangle_bnd.Boundary()
    target = DomainTypeArray([(IF, IF, NEG), (NEG, IF, IF), (IF, NEG, IF)])
    assert triangle_bnd_bnd == target
    assert triangle_bnd_bnd.codim == 2

    lines = []
    for i in range(nr_ls):
        line_i = tuple(IF if ii == i else NEG for ii in range(nr_ls))
        lines.append(DomainTypeArray(line_i))
        assert lines[i].codim == 1


# -----------------------------------------------------------------------------
# ------------------------------ TEST OPERATORS -------------------------------
# -----------------------------------------------------------------------------
def test_operators():
    dta1 = DomainTypeArray((POS, NEG))
    dta2 = DomainTypeArray((POS, POS))
    dta3 = dta1 | dta2

    assert (POS, NEG) in dta3
    assert (POS, POS) in dta3
    assert len(dta3) == 2

    target = DomainTypeArray([(POS, NEG), (POS, POS)])
    assert dta3 == target

    dta3 |= DomainTypeArray([(NEG, POS)])
    target = DomainTypeArray([(POS, NEG), (POS, POS), (NEG, POS)])
    assert dta3 == target

    dta4 = dta3 & dta1

    target = DomainTypeArray([(POS, NEG)])
    assert dta4 == target

    dta3 &= dta1 | dta2
    target = DomainTypeArray([(POS, NEG), (POS, POS)])
    assert dta3 == target

    dta5 = DomainTypeArray([(IF, NEG)]) & DomainTypeArray([(POS, NEG)])
    target = DomainTypeArray([(IF, NEG)])
    assert dta5 == target

    dta5 &= DomainTypeArray([(POS, IF)])
    target = DomainTypeArray([(IF, IF)])
    assert dta5 == target

    target = DomainTypeArray([(NEG, POS), (NEG, NEG)])
    assert (~dta3) == target

    dta6 = DomainTypeArray([(IF, NEG, POS)])
    dta6 &= DomainTypeArray([(POS, IF, POS)])
    target = DomainTypeArray([(IF, NEG, IF), (POS, IF, IF), (IF, POS, IF),
                              (NEG, IF, IF), (IF, IF, NEG)])
    assert (~dta6) == target

    dta7 = DomainTypeArray([(NEG, NEG, NEG)])
    target = DomainTypeArray([(NEG, IF, NEG), (NEG, NEG, IF), (IF, NEG, NEG)])
    assert dta7.Boundary() == target

    dta8 = DomainTypeArray([(POS, IF, NEG)])
    target = DomainTypeArray([(POS, IF, IF), (IF, IF, NEG)])
    assert dta8.Boundary() == target


# -----------------------------------------------------------------------------
# ---------------------------- TEST ANY EXPANSION -----------------------------
# -----------------------------------------------------------------------------
def test_any_expansion():
    dta9 = DomainTypeArray((NEG, ANY, ANY))
    target = DomainTypeArray([(NEG, NEG, NEG), (NEG, NEG, POS), 
                            (NEG, POS, NEG), (NEG, POS, POS)])
    assert dta9 == target

    dta10 = DomainTypeArray([(NEG, ANY, ANY), (POS, ANY, ANY)])
    dta11 = DomainTypeArray([(ANY, ANY, ANY)])
    assert dta10 == dta11

    dta12 = DomainTypeArray([(IF, ANY, ANY), (POS, IF, ANY)])
    target = DomainTypeArray([(IF, NEG, NEG), (IF, NEG, POS), (IF, POS, NEG), 
                              (IF, POS, POS), (POS, IF, NEG), (POS, IF, POS)])
    assert dta12 == target


# -----------------------------------------------------------------------------
# -------------------------- TEST TENSOR OPERATORS ----------------------------
# -----------------------------------------------------------------------------
def test_tensor_operators():
    dta1 = DomainTypeArray((POS,))
    dta2 = DomainTypeArray([(POS, NEG), (POS,POS)])
    dta3 = DomainTypeArray((NEG, NEG, IF))

    union = TensorUnion(dta1, dta2)
    target = DomainTypeArray([(POS, ANY, ANY), (ANY, POS, NEG), (ANY, POS, POS)])
    assert union == target

    with pytest.raises(Exception):
        union = TensorUnion(dta1, dta2, dta3)

    intersection1 = TensorIntersection(dta1, dta2)
    target = DomainTypeArray([(POS, POS, NEG), (POS, POS, POS)])
    assert intersection1 == target

    intersection2 = TensorIntersection(dta1, dta2, dta3)
    target = DomainTypeArray([(POS, POS, NEG, NEG, NEG, IF), 
                              (POS, POS,POS, NEG, NEG, IF)])
    assert intersection2 == target


# -----------------------------------------------------------------------------
# --------------------------- RUN TESTS SEPARATELY ----------------------------
# -----------------------------------------------------------------------------
if __name__ == "__main__":
    test_trianlge()
    test_operators()
    test_any_expansion()
    test_tensor_operators()



# unfitted Heat equation with Neumann b.c.
from ngsolve import *
from time import sleep
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters

from ngsolve.internal import *
from xfem import *
from numpy import pi

from xfem.lset_spacetime import *

square = SplineGeometry()
square.AddRectangle([-1,-0.75],[1,1.5],bc=1)
ngmesh = square.GenerateMesh(maxh=10, quad_dominated=True)
mesh = Mesh (ngmesh)

for i in range(6):
    mesh.Refine()

fes1 = H1(mesh, order=1, dirichlet=[])
fes_lset_slice = H1(mesh, order=1, dirichlet=[])
k_t = 1
k_s = 1
lset_order_time = 2
tfe = ScalarTimeFE(k_t) 

st_fes = SpaceTimeFESpace(fes1,tfe)
visoptions.deformation = 1

#Fitted heat equation example
tend = 0.5
delta_t = 1/32
tnew = 0

told = Parameter(0)
tref = ReferenceTimeVariable()
t = told + delta_t*tref

lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = k_s, order_time = lset_order_time,
                                         threshold=0.5, discontinuous_qn=True)

r0 = 0.5
tmp1 = 0.5*pi/r0                                         
rho =  CoefficientFunction((1/(pi))*sin(2*pi*t))
d_rho = CoefficientFunction(2*cos(2*pi*t))
w = CoefficientFunction((0,d_rho)) # convection
levelset= CoefficientFunction(sqrt(x*x+(y-rho)*(y-rho)) -r0)
coeff_f = CoefficientFunction(  -(pi/r0)*tmp1*( sin(tmp1*sqrt(x*x+(y-rho)*(y-rho)))*sin(tmp1*sqrt(x*x+(y-rho)*(y-rho))) 
                                - cos(tmp1*sqrt(x*x+(y-rho)*(y-rho)))*cos(tmp1*sqrt(x*x+(y-rho)*(y-rho))) )  
                                + (pi/r0)*cos(tmp1*sqrt(x*x+(y-rho)*(y-rho)))*sin(tmp1*sqrt(x*x+(y-rho)*(y-rho)))*(1/sqrt(x*x+(y-rho)*(y-rho))) )

# for monitoring the error
u_exact = CoefficientFunction(cos(tmp1*sqrt(x*x+(y-rho)*(y-rho)))*cos(tmp1*sqrt(x*x+(y-rho)*(y-rho))) )


u0 = GridFunction(st_fes)
u0_ic = GridFunction(fes1)
Draw(u0_ic,mesh,"u")
u = st_fes.TrialFunction()
v = st_fes.TestFunction()

lset_top = GridFunction(fes_lset_slice)
lset_bottom = GridFunction(fes_lset_slice)

t_old = 0
u0_ic.Set(u_exact)


while tend - t_old > delta_t/2:
    
    dfm = lset_adap_st.CalcDeformation(levelset,told,t_old,delta_t) 
    lset_p1 = lset_adap_st.lset_p1    
    hasneg_spacetime = lset_adap_st.hasneg_spacetime
    active_dofs = GetDofsOfElements(st_fes,hasneg_spacetime)
    #print(active_dofs)
    
    lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
    # Put this into LevelSetMeshAdaptation_Spacetime later on
    lset_bottom.vec[:] = lset_p1.vec[0:fes_lset_slice.ndof]
    lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
    lset_top.vec[:] = lset_p1.vec[lset_order_time*fes_lset_slice.ndof : (lset_order_time+1)*fes_lset_slice.ndof]
    lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}
    
    a = BilinearForm(st_fes,symmetric=False)
    a += SymbolicBFI(levelset_domain = lset_neg, form = -u*dt(v), time_order=2)
    a += SymbolicBFI(levelset_domain = lset_neg, form = delta_t*grad(u)*grad(v), time_order=2)
    a += SymbolicBFI(levelset_domain = lset_neg, form = -delta_t*u*w*grad(v), time_order=2)
    a += SymbolicBFI(levelset_domain = lset_neg_top, form = fix_t(u,1)*fix_t(v,1) )
    
    #Draw(lset_bottom,mesh,"bottom") 
    #Draw(lset_top,mesh,"top") 
    #input("lvlsets")
    a.Assemble()
              
    f = LinearForm(st_fes)
    f += SymbolicLFI(levelset_domain = lset_neg, form = delta_t*coeff_f*v, time_order=2)
    f += SymbolicLFI(levelset_domain = lset_neg_bottom,form = u0_ic*fix_t(v,0) )
    f.Assemble()
    u0.vec.data = a.mat.Inverse(active_dofs,"") * f.vec
       
    # exploiting the nodal property of the time fe:
    u0_ic.vec[:] = u0.vec[fes1.ndof : 2*fes1.ndof]
    
    t_old = t_old + delta_t
    told.Set(t_old)
    
    l2error = sqrt(Integrate(lset_neg_top,(u_exact-u0_ic)*(u_exact-u0_ic),mesh))
    print("t = {0}, l2error = {1}".format(t_old,l2error))
   
    Redraw(blocking=True)
    #Draw(sqrt((u_exact-u0_ic)*(u_exact-u0_ic)),mesh,"error") 
       



from sympy import *
from sympy.printing import latex, pretty
from sympy.parsing.sympy_parser import parse_expr

r0 = 1
Q = pi/r0

def get_Laplace_u_str(rhostr, ustr):
    x,y,z,t = symbols("x y z t")
    
    rho = eval(rhostr) #(1 - y**2)*t
    r = sqrt((x - rho)**2+y**2+z**2)
    
    u = eval(ustr) #cos(Q*r) * sin(pi*t)
    Laplaceu = diff(u,x,2) + diff(u,y,2) + diff(u,z,2)
    
    #print("Laplace: ", Laplaceu)
    return str(Laplaceu)
    #Laplaceu = - Q*sin(pi*t)*(cos(Q*r)*Q* ( drdx**2 + drdy**2) + sin(Q*r)*( dr2dx2 + dr2dy2 ))
    #coeff_f = -Laplaceu + pi * cos(Q*r) * cos(pi*t)

#print ( get_Laplace_u_str("(1 - y**2)*t", "cos(Q*r) * sin(pi*t)"))

def get_dt_rho(rhostr):
    x,y,z,t = symbols("x y z t")
    
    rho = eval(rhostr)
    return str(diff(rho, t))



from ngsolve import *
from xfem import *
from xfem.mlset import *
from netgen.geom2d import unit_square
from netgen.csg import unit_cube
import sys
import time

ngsglobals.msg_level = 1

def test_fes_timing(dimension=2,stdfes=True,quad_dominated=False, order=1):
    if dimension == 2:
        mesh = Mesh (unit_square.GenerateMesh(maxh=0.2,quad_dominated=quad_dominated))
    else:
        mesh = Mesh (unit_cube.GenerateMesh(maxh=0.2,quad_dominated=quad_dominated))

    Vhs = H1(mesh, order=order, dirichlet=[1,2,3,4])

    lsetp1 = GridFunction(H1(mesh,order=1))
    InterpolateToP1((sqrt(sqrt(x*x+y*y)) - 1.0),lsetp1)

    if stdfes:
        Vh = Vhs
    else:
        Vhx = XFESpace(Vhs, lsetp1)
        Vh = Vhx
    container = Vh.__timing__()
    ts = time.time()
    steps = 5
    for i in range(steps):
        Vh.Update()
    te = time.time()
    return 1e9*(te-ts)/steps

testcases = [ (2, False, 1, False),
              (2,  True, 1, False),
              (3, False, 1, False),
              (3,  True, 1, False),
              (2, False, 3, False),
              (2,  True, 3, False),
              (3, False, 1, True),
              (3,  True, 1, True),
]

metricname_base = "fes_timings"
for dim, stdfes, order, taskmanager in testcases:
    metricname = metricname_base + "_dim" + str(dim)
    if stdfes:
        metricname += "_std"
    else:
        metricname += "_x"
    metricname += "_order" + str(order)
    if taskmanager:
        metricname += "_TM"
    else:
        metricname += "_NoTM"
    if taskmanager:
        with TaskManager():
            report = test_fes_timing(dimension=dim,stdfes=stdfes,order=order)
    else:
        report = test_fes_timing(dimension=dim,stdfes=stdfes,order=order)
    print(metricname,report, file=sys.stderr)


def dta_timings(n, steps):
    domains = [DomainTypeArray((NEG, NEG, NEG)) for _ in range(n)]

    ts1 = time.time()
    for i in range(steps):
        dta = TensorUnion(*domains)
    te1 = time.time()

    ts2 = time.time()
    for i in range(steps):
        dta_inv = ~dta
    te2 = time.time()

    return 1e9 * (te1 - ts1) / steps, len(dta.as_list),\
           1e9 * (te2 - ts2) / steps, len(dta_inv.as_list)


testcases = [1, 2, 3, 4, 5]
metricname_base = "dta_timings"

for n in testcases:
    metricname = metricname_base + "_tris" + str(n)
    report = dta_timings(n, 5)
    print(metricname, *report, file=sys.stderr)



# unfitted Heat equation with Neumann b.c.
# solved with a P1-DG-in-time space-time discretization
from ngsolve import *
from netgen.csg import *

from xfem import *
from math import pi

from xfem.lset_spacetime import *
from time import sleep

ngsglobals.msg_level = 1

cube = CSGeometry()
cube.Add (OrthoBrick(Pnt(-1,-1,-1), Pnt(1,1,1)))
ngmesh = cube.GenerateMesh(maxh=0.1, quad_dominated=False)
mesh = Mesh (ngmesh)

coef_told = Parameter(0)
coef_delta_t = Parameter(0)
tref = ReferenceTimeVariable()
t = coef_told + coef_delta_t*tref

r0 = 0.5

# position shift of the geometry in time
rho = CoefficientFunction(0.0) #CoefficientFunction((1/(pi))*sin(2*pi*t))
rhoL = lambda t: CoefficientFunction(0.0) #CoefficientFunction((1/(pi))*sin(2*pi*t))
#convection velocity:
d_rho = CoefficientFunction(0.0) #CoefficientFunction(2*cos(2*pi*t))
w = CoefficientFunction((0,d_rho,0))

# level set
r = sqrt(x**2+(y-rho)**2+z**2)
levelset= r - r0

# diffusion coefficient
alpha = 1

# solution and r.h.s.
Q = pi/r0   
u_exact = cos(Q*r) * sin(pi*t)
u_exactL = lambda t: cos(Q*sqrt(x**2+(y-rhoL(t))**2 + z**2)) * sin(pi*t)
coeff_f = (2*Q/r * sin(Q*r) + (Q**2) * cos(Q*r)) * sin(pi*t) + pi * cos(Q*r) * cos(pi*t)
u_init = u_exact

# polynomial order in time
k_t = 1
# polynomial order in space
k_s = 1
# spatial FESpace for solution
fes1 = H1(mesh, order=k_s)
# polynomial order in time for level set approximation
lset_order_time = 1
# integration order in time
time_order = 2
# time finite element (nodal!)
tfe = ScalarTimeFE(k_t)
# space-time finite element space
st_fes = SpaceTimeFESpace(fes1,tfe, flags = {"dgjumps": True})

#Unfitted heat equation example
tend = 0.5
delta_t = tend/16
coef_delta_t.Set(delta_t)
tnew = 0
told = 0

lset_p1 = GridFunction(st_fes)

SpaceTimeInterpolateToP1(levelset,tref,lset_p1)

lset_top = CreateTimeRestrictedGF(lset_p1,1.0)
lset_bottom = CreateTimeRestrictedGF(lset_p1,0.0)

gfu = GridFunction(st_fes)

u_last = CreateTimeRestrictedGF(gfu,0)
u_last.Set(u_exactL(0.))

u,v = st_fes.TnT()

h = specialcf.mesh_size

lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}

def SpaceTimeNegBFI(form):
    return SymbolicBFI(levelset_domain = lset_neg, form = form, time_order=time_order)

ci = CutInfo(mesh,time_order=time_order)
n_F = specialcf.normal(mesh.dim)

hasneg_integrators_a = []
hasneg_integrators_f = []
patch_integrators_a = []
hasneg_integrators_a.append(SpaceTimeNegBFI(form = delta_t*alpha*grad(u)*grad(v)))
hasneg_integrators_a.append(SymbolicBFI(levelset_domain = lset_neg_top, form = fix_t(u,1)*fix_t(v,1)))
hasneg_integrators_a.append(SpaceTimeNegBFI(form = -u*dt(v)))
hasneg_integrators_a.append(SpaceTimeNegBFI(form = -delta_t*u*InnerProduct(w,grad(v))))

#patch_integrators_a.append(SymbolicFacetPatchBFI(20*h*InnerProduct(grad(u) - grad(u.Other()), n_F) * InnerProduct(grad(v) - grad(v.Other()), n_F), skeleton=True, time_order=time_order))

patch_integrators_a.append(SymbolicFacetPatchBFI(form = delta_t*1.05*h**(-2)*(u-u.Other())*(v-v.Other()), skeleton=False, time_order=time_order))

hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg, form = delta_t*coeff_f*v, time_order=time_order)) 
hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg_bottom,form = u_last*fix_t(v,0)))


a = BilinearForm(st_fes,check_unused=False,symmetric=False)
for integrator in hasneg_integrators_a + patch_integrators_a:
    a += integrator

f = LinearForm(st_fes)

for integrator in hasneg_integrators_f:
    f += integrator

while tend - told > delta_t/2:
    SpaceTimeInterpolateToP1(levelset,tref,lset_p1)
    RestrictGFInTime(spacetime_gf=lset_p1,reference_time=0.0,space_gf=lset_bottom)
    RestrictGFInTime(spacetime_gf=lset_p1,reference_time=1.0,space_gf=lset_top)
    
    # update markers in (space-time) mesh
    ci.Update(lset_p1,time_order=time_order)

    # re-compute the facets for stabilization:
    ba_facets = GetFacetsWithNeighborTypes(mesh,a=ci.GetElementsOfType(HASNEG),
                                                b=ci.GetElementsOfType(IF))
    # re-evaluate the "active dofs" in the space time slab
    active_dofs = GetDofsOfElements(st_fes,ci.GetElementsOfType(HASNEG))
    
    # re-set definedonelements-markers according to new markings:
    for integrator in hasneg_integrators_a + hasneg_integrators_f:
        integrator.SetDefinedOnElements(ci.GetElementsOfType(HASNEG))
    for integrator in patch_integrators_a:
        integrator.SetDefinedOnElements(ba_facets)
    
    # assemble linear system
    a.Assemble()
    f.Assemble()
    
    # solve linear system
    inv = a.mat.Inverse(active_dofs,inverse="")
    gfu.vec.data =  inv * f.vec

    # evaluate upper trace of solution for
    #  * for error evaluation 
    #  * upwind-coupling to next time slab
    RestrictGFInTime(spacetime_gf=gfu,reference_time=1.0,space_gf=u_last)   

    # update time variable (float and ParameterCL)
    told = told + delta_t
    coef_told.Set(told)
    
    # compute error at end of time slab
    l2error = sqrt(Integrate(lset_neg_top,(u_exactL(told) -u_last)**2,mesh))
    # print time and error
    print("t = {0:10}, l2error = {1:20}".format(told,l2error),end="\n")



"""
    Description:
    -----------
    Zalesak Disk described my multiple level sets. Inspired by [1].

    References:
    -----------
    [1] D. P. Starinshak, S. Karni and P. L. Roe: A New Level-Set
        Model for the Representation of Non-Smooth Geometries. In J Sci 
        Comput (2014) 61(3):649-672. DOI:10.1007/s10915-014-9842-0

"""

from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *
from math import pi, asin, sqrt
import pytest

SetNumThreads(4)

geo = SplineGeometry()
geo.AddRectangle((-1.1, -1.1), (1.1, 1.1), bc=1)

with TaskManager():
    mesh = Mesh(geo.GenerateMesh(maxh=0.1))

level_sets = [x * x + y * y - 1, -x - 1 / 3, x - 1 / 3, y - 0.5]
nr_ls = len(level_sets)
level_sets_p1 = tuple(GridFunction(H1(mesh, order=1)) for i in range(nr_ls))

for i, lsetp1 in enumerate(level_sets_p1):
    InterpolateToP1(level_sets[i], lsetp1)
    Draw(lsetp1, mesh, "lset_p1_" + str(i))

SetVisualization(min=0, max=0)
Redraw()

area_zd = pi - 2 / 3 * (0.5 + sqrt(2) * 2 / 3)
area_zd += - (2 * asin(1 / 3) - sin(2 * asin(1 / 3))) / 2
print("Area   = {:12.10f}".format(area_zd))


z_disc1 = DomainTypeArray([(NEG, NEG, NEG, POS), (NEG, POS, NEG, POS),
                           (NEG, POS, NEG, NEG), (NEG, NEG, POS, NEG),
                           (NEG, NEG, POS, POS)])
Draw(z_disc1.Indicator(level_sets_p1), mesh, "z_disc1")

lset_zdisc1 = {"levelset": level_sets_p1, "domain_type": z_disc1}

area1 = Integrate(levelset_domain=lset_zdisc1, cf=1, mesh=mesh, order=0)
print("Area 1 = {:12.10f}".format(area1))
print("Error: {:4.2e}".format(abs(area1 - area_zd)))


# Alternative 1
z_disc2 = DomainTypeArray([(NEG, ANY, ANY, POS), (NEG, POS, ANY, ANY),
                           (NEG, ANY, POS, ANY)])
# print(z_disc2)
Draw(z_disc2.Indicator(level_sets_p1), mesh, "z_disc2")
z_disc2.Compress(level_sets_p1, persistent=True)
print(z_disc2==z_disc1)
bnd2 = z_disc2.Boundary()
Draw(z_disc2.Indicator(level_sets_p1), mesh, "z_disc2_c")
lset_zdisc2 = {"levelset": level_sets_p1, "domain_type": z_disc2}

area2 = Integrate(levelset_domain=lset_zdisc2, cf=1, mesh=mesh, order=0)
print("Area 2 = {:12.10f}".format(area2))
print("Error: {:4.2e}".format(abs(area2 - area_zd)))
assert abs(area1 - area2) < 1e-12

# # Alternative 2
part1 = DomainTypeArray((NEG,ANY,ANY,ANY))
part2 = DomainTypeArray((NEG,NEG,NEG,NEG))
z_disc3 = part1 & ~part2

lset_zdisc3 = {"levelset": level_sets_p1, "domain_type": z_disc3}

area3 = Integrate(levelset_domain=lset_zdisc3, cf=1, mesh=mesh, order=0)
print("Area 3 = {:12.10f}".format(area3))
print("Error: {:4.2e}".format(abs(area3 - area_zd)))
assert abs(area1 - area3) < 1e-12


bnd_dta = DomainTypeArray([(IF, NEG, NEG, POS), (IF, NEG, POS, POS), 
                           (IF, NEG, POS, NEG), (NEG, NEG, IF, NEG), 
                           (NEG, NEG, NEG, IF), (NEG, IF, NEG, NEG),
                           (IF, POS, NEG, NEG), (IF, POS, NEG, POS)])

bnd3 = z_disc3.Boundary()
bnd3.Compress(level_sets_p1)

assert bnd3 == bnd_dta
assert bnd2 == bnd3

Draw(bnd_dta.IndicatorSmoothed(level_sets_p1), mesh, "bnd", sd=6)

normals = z_disc2.GetOuterNormals(level_sets_p1)

for dtt, n in normals.items():
    Draw(n, mesh, "normal")
    Draw(DomainTypeArray(dtt).IndicatorSmoothed(level_sets_p1), mesh, "boundary")
    # input(dtt.__str__())

# Test operators
dta_t1 = bnd3 & bnd2
assert dta_t1.lsets == None
assert dta_t1.persistent_compress == False
dta_t2 = bnd3 | bnd2
assert dta_t2.lsets == None
assert dta_t2.persistent_compress == False

bnd4 = z_disc3.Boundary()
bnd4.Compress(level_sets_p1, persistent=True)

dta_t3 = bnd4 & bnd2
assert dta_t3.lsets == level_sets_p1
assert dta_t3.persistent_compress == True

dta_t3 = bnd4 | bnd2
assert dta_t3.lsets == level_sets_p1
assert dta_t3.persistent_compress == True

level_sets_p1_2 = tuple(GridFunction(H1(mesh, order=1)) for i in range(nr_ls))
for i, lsetp1 in enumerate(level_sets_p1_2):
    InterpolateToP1(level_sets[i], lsetp1)
    Draw(lsetp1, mesh, "lset_p1_" + str(i))

bnd5 = z_disc3.Boundary()
bnd5.Compress(level_sets_p1_2, persistent=True)

# These operations should raise an exception
with pytest.raises(Exception):
    dta_t4 = bnd5 & z_disc2.Boundary()
with pytest.raises(Exception):
    dta_t4 = bnd5 | z_disc2.Boundary()


# integration on lset domains
from math import pi
# ngsolve stuff
from ngsolve import *
# basic xfem functionality
from xfem import *
# For LevelSetAdaptationMachinery
from xfem.lsetcurv import *

from ngsolve.meshes import *

import pytest

@pytest.mark.parametrize("quad", [False,True])
@pytest.mark.parametrize("order", [1,2,3])

def test_intcurved(quad, order):
    levelset = sqrt(x*x+y*y)-0.5
    referencevals = { POS : 4.0-0.25*pi, NEG : 0.25*pi, IF : pi }

    N=4
    errors_uncurved = dict()
    errors_curved = dict()
    eoc_uncurved = dict()
    eoc_curved = dict()

    for key in [NEG,POS,IF]:
        errors_curved[key] = []
        errors_uncurved[key] = []
        eoc_curved[key] = []
        eoc_uncurved[key] = []

    refinements = 5
    if order == 1:
        refinements += 2
    for reflevel in range(refinements):

        mesh = MakeStructured2DMesh(quads = quad, nx=N*2**reflevel, ny=N*2**reflevel,
                                    mapping = lambda x,y : (2*x-1,2*y-1))    
        lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.2, discontinuous_qn=True)
        lsetp1 = lsetmeshadap.lset_p1

        f = CoefficientFunction (1.0)

        for key in [NEG,POS,IF]:
            # Applying the mesh deformation
            deformation = lsetmeshadap.CalcDeformation(levelset)

            integrals_uncurved = Integrate(levelset_domain = { "levelset" : lsetp1, "domain_type" : key},
                                           cf=f, mesh=mesh, order = order)

            mesh.SetDeformation(deformation)
            integrals_curved = Integrate(levelset_domain = { "levelset" : lsetp1, "domain_type" : key},
                                           cf=f, mesh=mesh, order = order)
            # Unapply the mesh deformation (for refinement)
            mesh.UnsetDeformation()

            errors_curved[key].append(abs(integrals_curved - referencevals[key]))
            errors_uncurved[key].append(abs(integrals_uncurved - referencevals[key]))
        # refine cut elements:
        # if not quad:
        #     RefineAtLevelSet(gf=lsetmeshadap.lset_p1)

    for key in [NEG,POS,IF]:
        eoc_curved[key] = [log(a/b)/log(2) for (a,b) in zip (errors_curved[key][0:-1],errors_curved[key][1:]) ]
        eoc_uncurved[key] = [log(a/b)/log(2) for (a,b) in zip (errors_uncurved[key][0:-1],errors_uncurved[key][1:]) ]

    # print("errors (uncurved):  \n{}\n".format(errors_uncurved))
    # print("   eoc (uncurved):  \n{}\n".format(   eoc_uncurved))
    print("errors (  curved):  \n{}\n".format(  errors_curved))
    print("   eoc (  curved):  \n{}\n".format(     eoc_curved))

    print("avg.eoc(curved, IF):  \n{}\n".format(     sum(eoc_curved[IF][2:])/len(eoc_curved[IF][2:])))
    print("avg.eoc(curved,NEG):  \n{}\n".format(     sum(eoc_curved[NEG][2:])/len(eoc_curved[NEG][2:])))
    print("avg.eoc(curved,POS):  \n{}\n".format(     sum(eoc_curved[POS][2:])/len(eoc_curved[POS][2:])))

    if (order > 1):
        assert errors_curved[IF][-1] < 1e-5
        assert errors_curved[NEG][-1] < 1e-5
        assert errors_curved[POS][-1] < 1e-5
    else:
        assert errors_curved[IF][-1] < 1e-4
        assert errors_curved[NEG][-1] < 1e-4
        assert errors_curved[POS][-1] < 1e-4

    s = 0
    if order == 1:
        s+=2
    assert sum(eoc_curved[IF][s:])/len(eoc_curved[IF][s:]) > order + 0.75
    assert sum(eoc_curved[NEG][s:])/len(eoc_curved[NEG][s:]) > order + 0.75
    assert sum(eoc_curved[POS][s:])/len(eoc_curved[POS][s:]) > order + 0.75



from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

square = SplineGeometry()
square.AddRectangle((-4, -3), (4, 3), bc=1)
mesh = Mesh(square.GenerateMesh(maxh=0.5, quad_dominated=False))


level_sets = [-y+1, 2*x-y, -2*x-y]  

# level_sets = [x - 3, -x - y + 1, -x + y + 1,
#               -x - 3, x - y + 1, x + y + 1 ]

level_sets_p1 = tuple(GridFunction(H1(mesh, order=1))
                      for i in range(len(level_sets)))
for i, lsetp1 in enumerate(level_sets_p1):
    InterpolateToP1(level_sets[i], lsetp1)
    Draw(lsetp1, mesh, "lsetp1_{}".format(i))



triangle = DomainTypeArray((POS, NEG, NEG))
normals = triangle.GetOuterNormals(level_sets_p1)

for dtt in triangle.Boundary():
    name = str(dtt)
    for s in ["DOMAIN_TYPE.", ",", " ", "(", ")"]:
        name = name.replace(s, "")
    Draw(normals[(POS, NEG, NEG)][dtt], mesh, "normals_"+name)

Draw(triangle.Indicator(level_sets), mesh, "triangle")











# triangles = DomainTypeArray([(NEG, NEG, NEG, ANY, ANY, ANY),
#                              (ANY, ANY, ANY, NEG, NEG, NEG)])

# print(triangles)

# print(triangles.Boundary())



# %% ------------------------- LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

ngsglobals.msg_level = 2

# %% --------------------------- PARAMETERS ---------------------------------
# Domain corners
ll, ur = (0, -0.5), (2, 0.5)
# Initial mesh diameter
initial_maxh = 0.4
# Number of mesh bisections
nref = 4
# Order of finite element space
k = 1

# Stabilization parameter for ghost-penalty
gamma_s = 0.5
# Stabilization parameter for Nitsche
gamma_n = 10

# ----------------------------------- MAIN ------------------------------------
# %% Set up the level sets, exact solution and right-hand side


def level_sets():
    return [x - 1.5, y - 0.15, -0.15 - y]


nr_ls = len(level_sets())
rhs = CF((0, -0.1))

# %% Geometry and mesh
geo = SplineGeometry()
geo.AddRectangle(ll, ur, bcs=("bottom", "right", "top", "left"))
ngmesh = geo.GenerateMesh(maxh=initial_maxh)
for i in range(nref):
    ngmesh.Refine()
mesh = Mesh(ngmesh)


# %% Level set and cut-information
P1 = H1(mesh, order=1)
lsetsp1 = tuple(GridFunction(P1) for i in range(nr_ls))
for i, lsetp1 in enumerate(lsetsp1):
    InterpolateToP1(level_sets()[i], lsetp1)
    Draw(lsetp1, mesh, "lsetp1_{}".format(i))

# %%
square = DomainTypeArray((NEG, NEG, NEG))
with TaskManager():
    square.Compress(lsetsp1)
    boundary = square.Boundary()
    boundary.Compress(lsetsp1)

mlci = MultiLevelsetCutInfo(mesh, lsetsp1)

# %%
# Element and degrees-of-freedom markers
els_if_singe = {dtt: BitArray(mesh.ne) for dtt in boundary}
facets_gp = BitArray(mesh.nedge)

hasneg = mlci.GetElementsWithContribution(square)
Draw(BitArrayCF(hasneg), mesh, "hasneg")

# %% Finite element space
Vhbase = VectorH1(mesh, order=k, dirichlet="left", dgjumps=True)
Vh = Restrict(Vhbase, hasneg)
gfu = GridFunction(Vh)

hasif = mlci.GetElementsWithContribution(boundary)
Draw(BitArrayCF(hasif), mesh, "hasif")

for i, (dtt, els_bnd) in enumerate(els_if_singe.items()):
    els_bnd[:] = mlci.GetElementsWithContribution(dtt)
    Draw(BitArrayCF(els_bnd), mesh, "els_if_singe" + str(i))

facets_gp = GetFacetsWithNeighborTypes(mesh, a=hasneg, b=hasif,
                                       use_and=True)

els_gp = GetElementsWithNeighborFacets(mesh, facets_gp)
Draw(BitArrayCF(els_gp), mesh, "gp_elements")

# %% Bilinear and linear forms of the weak formulation
u, v = Vh.TnT()
h = specialcf.mesh_size
normals = square.GetOuterNormals(lsetsp1)

# Set up the integrator symbols
dx = dCut(lsetsp1, square, definedonelements=hasneg)
ds = {dtt: dCut(lsetsp1, dtt, definedonelements=els_if_singe[dtt])
      for dtt in boundary}
dw = dFacetPatch(definedonelements=facets_gp)


def eps:
    return Sym(Grad(u))


E, nu = 210, 0.2
mu = E / 2 / (1+nu)
lam = E * nu / ((1+nu)*(1-2*nu))

# Construct integrator
a = RestrictedBilinearForm(Vh, facet_restriction=facets_gp, check_unused=False)
a += 2 * mu * InnerProduct(eps(u), eps(v)) * dx + lam * div(u) * div(v) * dx
a += mu * gamma_s / (h**2) * (u - u.Other()) * (v - v.Other()) * dw

f = LinearForm(Vh)
f += rhs * v * dx


# Assemble and solve the linear system
f.Assemble()
a.Assemble()

gfu.vec.data = a.mat.Inverse(Vh.FreeDofs()) * f.vec

# %% Draw webgui
ind = IfPos(lsetsp1[0], 1, IfPos(lsetsp1[1], 1, IfPos(lsetsp1[2], 1, -1)))
DrawDC(ind, gfu[1], 0, mesh, "u")

# %% pyvista vis
try:
    import pyvista
    vtk = VTKOutput(mesh, coefs=[*lsetsp1, 33*CF((gfu[0], gfu[1], 0))],
                    names=["lset1", "lset2", "lset3", "u"], filename="vtkout",
                    subdivision=3, floatsize="single", legacy=False).Do()
    pyvista.set_jupyter_backend('client')
    visobj = pyvista.read('vtkout.vtu')
    plot = pyvista.Plotter()
    arrows = visobj.glyph(scale="u", orient="u", factor=0.15, tolerance=0.03)
    clipped0 = visobj.clip_scalar("lset1", invert=True) \
                     .clip_scalar("lset2", invert=True) \
                     .clip_scalar("lset3", invert=True)
    visobj2 = visobj.warp_by_vector(factor=0.15)
    clipped = visobj2.clip_scalar("lset1", invert=True) \
                     .clip_scalar("lset2", invert=True) \
                     .clip_scalar("lset3", invert=True)
    plot.add_mesh(clipped0, label='reference conf.')
    plot.add_mesh(clipped, label='deformed conf.')
    plot.add_mesh(arrows, label='deform. direction')
    plot.add_legend()
    plot.show()
except ImportError:
    print("pyvista not installed")
    pass



"""
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.meshing import Element2D, Element3D, MeshPoint, FaceDescriptor
from netgen.meshing import Mesh as NetMesh
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from ngsolve import Mesh as NGSMesh


# ----------------------------------- MAIN ------------------------------------
def MakeUniform3DGrid(quads=False, N=5, P1=(0, 0, 0), P2=(1, 1, 1)):

    if not quads:
        raise BaseException("Only hex cube so far...")

    Lx = P2[0] - P1[0]
    Ly = P2[1] - P1[1]
    Lz = P2[2] - P1[2]

    cube = OrthoBrick(Pnt(P1[0], P1[1], P1[2]), Pnt(P2[0], P2[1], P2[2])).bc(1)
    geom = CSGeometry()
    geom.Add(cube)
    netmesh = NetMesh()
    netmesh.SetGeometry(geom)
    netmesh.dim = 3

    pids = []
    for i in range(N + 1):
        for j in range(N + 1):
            for k in range(N + 1):
                pids.append(netmesh.Add(MeshPoint(Pnt(P1[0] + Lx * i / N,
                                                      P1[1] + Ly * j / N,
                                                      P1[2] + Lz * k / N))))

    for i in range(N):
        for j in range(N):
            for k in range(N):
                base = i * (N + 1) * (N + 1) + j * (N + 1) + k
                baseup = base + (N + 1) * (N + 1)
                pnum = [base, base + 1, base + (N + 1) + 1, base + (N + 1),
                        baseup, baseup + 1, baseup + (N + 1) + 1,
                        baseup + (N + 1)]
                elpids = [pids[p] for p in pnum]
                netmesh.Add(Element3D(1, elpids))

    def AddSurfEls(p1, dx, dy, facenr):
        for i in range(N):
            for j in range(N):
                base = p1 + i * dx + j * dy
                pnum = [base, base + dx, base + dx + dy, base + dy]
                elpids = [pids[p] for p in pnum]
                netmesh.Add(Element2D(facenr, elpids))

    netmesh.Add(FaceDescriptor(surfnr=1, domin=1, bc=1))
    AddSurfEls(0, 1, N + 1, facenr=1)

    netmesh.Add(FaceDescriptor(surfnr=2, domin=1, bc=1))
    AddSurfEls(0, (N + 1) * (N + 1), 1, facenr=2)

    netmesh.Add(FaceDescriptor(surfnr=3, domin=1, bc=1))
    AddSurfEls(0, N + 1, (N + 1) * (N + 1), facenr=3)

    netmesh.Add(FaceDescriptor(surfnr=4, domin=1, bc=1))
    AddSurfEls((N + 1)**3 - 1, -(N + 1), -1, facenr=1)

    netmesh.Add(FaceDescriptor(surfnr=5, domin=1, bc=1))
    AddSurfEls((N + 1)**3 - 1, -(N + 1) * (N + 1), -(N + 1), facenr=1)

    netmesh.Add(FaceDescriptor(surfnr=6, domin=1, bc=1))
    AddSurfEls((N + 1)**3 - 1, -1, -(N + 1) * (N + 1), facenr=1)

    netmesh.Compress()
    mesh = NGSMesh(netmesh)
    mesh.ngmesh.Save("tmp.vol.gz")
    mesh = NGSMesh("tmp.vol.gz")
    return mesh



import pytest
from math import pi
from ngsolve import *
from netgen.geom2d import SplineGeometry
from xfem import *

@pytest.mark.parametrize("quad", [True, False])
def test_spacetime_vtk(quad):
  tref = ReferenceTimeVariable()
  square = SplineGeometry()
  A = 1.25
  square.AddRectangle([-A,-A],[A,A])
  ngmesh = square.GenerateMesh(maxh=0.3, quad_dominated=False)
  mesh = Mesh (ngmesh)
  
  #### expression for the time variable: 
  
  coef_told = Parameter(0)
  coef_delta_t = Parameter(0)
  tref = ReferenceTimeVariable()
  t = coef_told + coef_delta_t*tref
  
  #### the data: 
  # radius of disk (the geometry)
  r0 = 0.5
  x0 = lambda t: r0 * cos(pi*t)
  y0 = lambda t: r0 * sin(pi*t)
  
  #levelset= x - t
  levelset= sqrt((x-x0(t))**2+(y-y0(t))**2) - 0.4
  # spatial FESpace for solution
  fesp1 = H1(mesh, order=1)
  # polynomial order in time for level set approximation
  lset_order_time = 1
  # integration order in time
  time_order = 2
  # time finite element (nodal!)
  tfe = ScalarTimeFE(1) 
  # space-time finite element space
  st_fes = SpaceTimeFESpace(fesp1,tfe)
  
  coef_delta_t.Set(0.25)
  
  lset_p1 = GridFunction(st_fes)
  SpaceTimeInterpolateToP1(levelset,tref,lset_p1)
  
  vtk = SpaceTimeVTKOutput(ma=mesh,coefs=[levelset,lset_p1],
                           names=["levelset","lset_p1"],
                           filename="spacetime_vtk_test",
                           subdivision_x=1,subdivision_t=1)
  vtk.Do(t_start=coef_told.Get(), t_end=coef_told.Get() + coef_delta_t.Get())
  import os
  os.remove("spacetime_vtk_test_0.vtk")

if __name__ == "__main__":
    test_spacetime_vtk(False)
    test_spacetime_vtk(True)
    



import pytest
from ngsolve import *
from xfem import *
from netgen.geom2d import unit_square
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from ngsolve.meshes import MakeStructured2DMesh, MakeStructured3DMesh


@pytest.mark.parametrize("struc_mesh", [True, False])
@pytest.mark.parametrize("quad", [False])
@pytest.mark.parametrize("ROOTS", [POS, NEG])
@pytest.mark.parametrize("dim, levelset", [(2, x - 0.77654),
                                           (2, (x - 0.5)**2 + (y - 0.5)**2 - 0.3**2),
                                           (3, (x - 0.5)**2 + (y - 0.5)**2 + (z - 0.5)**2 - 0.3**2)
                                           ])
def test_aggregates(dim, struc_mesh, quad, ROOTS, levelset):
    if dim == 2:
        if struc_mesh:
            mesh = MakeStructured2DMesh(nx=20, ny=20, quads=quad)
        else:
            mesh = Mesh(unit_square.GenerateMesh(maxh=0.05,
                                                 quad_dominated=quad))
    if dim == 3:
        if struc_mesh:
            mesh = MakeStructured3DMesh(hexes=quad, nx=10)
        else:
            if quad:
                return None
            cube = CSGeometry()
            cube.Add(OrthoBrick(Pnt(-1, -1, -1), Pnt(1, 1, 1)))
            mesh = Mesh(cube.GenerateMesh(maxh=0.1, quad_dominated=quad))
    EA = ElementAggregation(mesh)

    gfu = GridFunction(H1(mesh))
    InterpolateToP1(levelset, gfu)

    ci = CutInfo(mesh, gfu)
    roots = ci.GetElementsOfType(ROOTS)
    bads = ci.GetElementsOfType(IF)

    EA.Update(roots, bads)

    # print("EA.GetInnerPatchFacets()", EA.GetInnerPatchFacets())
    els_surround_patch = GetElementsWithNeighborFacets(mesh, EA.patch_interior_facets)

    facets_gp = GetFacetsWithNeighborTypes(mesh,
                                           a=ci.GetElementsOfType(HAS(ROOTS)),
                                           b=ci.GetElementsOfType(IF),
                                           use_and=True, bnd_val_a=False,
                                           bnd_val_b=False)
    els_surround_gp = GetElementsWithNeighborFacets(mesh, facets_gp)

    Draw(BitArrayCF(els_surround_patch), mesh, "surrounding_facets")
    Draw(BitArrayCF(els_surround_gp), mesh, 'surround_gp')
    Draw(BitArrayCF(bads), mesh, 'if')
    Draw(BitArrayCF(els_surround_patch & ~els_surround_gp), mesh, 'bad_els')

    assert sum(els_surround_patch & ~els_surround_gp) == 0
    assert sum(els_surround_patch & ~ci.GetElementsOfType(HAS(ROOTS))) == 0

    return None


@pytest.mark.parametrize("struc_mesh", [True, False])
@pytest.mark.parametrize("quad", [True, False])
@pytest.mark.parametrize("nr", [10])
@pytest.mark.parametrize("dim, levelset", [(2, x - 0.77654),
                                           (2, (x - 0.5)**2 + (y - 0.5)**2 - 0.3**2),
                                           (3, (x - 0.5)**2 + (y - 0.5)**2 + (z - 0.5)**2 - 0.3**2)
                                           ])
def test_elements_with_contribution(dim, struc_mesh, quad, nr, levelset):
    if dim == 2:
        if struc_mesh:
            mesh = MakeStructured2DMesh(nx=20, ny=20, quads=quad)
        else:
            mesh = Mesh(unit_square.GenerateMesh(maxh=0.05,
                                                 quad_dominated=quad))
    if dim == 3:
        if struc_mesh:
            mesh = MakeStructured3DMesh(hexes=quad, nx=10)
        else:
            if quad:
                return None
            cube = CSGeometry()
            cube.Add(OrthoBrick(Pnt(-1, -1, -1), Pnt(1, 1, 1)))
            mesh = Mesh(cube.GenerateMesh(maxh=0.1, quad_dominated=quad))

    gfu = GridFunction(H1(mesh))
    InterpolateToP1(levelset, gfu)
    ci = CutInfo(mesh, gfu)

    for i in range(1, nr):
        els_test = ci.GetElementsWithThresholdContribution(NEG, i / nr)
        assert sum(els_test & ci.GetElementsOfType(POS)) == 0

        els_test = ci.GetElementsWithThresholdContribution(POS, i / nr)
        assert sum(els_test & ci.GetElementsOfType(NEG)) == 0

    return None

if __name__ == "__main__":
    test_aggregates(2, False, False, NEG, (x - 0.5)**2 + (y-0.5)**2 - 0.3**2)
    test_aggregates(3, False, False, NEG, (3, x - 0.77653))




from ngsolve import *
from xfem import *
from netgen.geom2d import SplineGeometry
from netgen.geom2d import unit_square

import pytest

@pytest.mark.parametrize("space", [H1,L2])
@pytest.mark.parametrize("order", [1,2,3])

def test_patch_local_solve(space,order):
    mesh = Mesh(unit_square.GenerateMesh(maxh=0.5))
    fes = space(mesh, order=order)

    EA = ElementAggregation(mesh)
    els_all, els_none = BitArray(mesh.ne), BitArray(mesh.ne)
    els_all.Set()
    els_none.Clear() 
    EA.Update(els_all, els_none)

    #print('non trivial patch elements\n', EA.els_in_nontrivial_patch)
    #print('elements in trivial patches\n', EA.els_in_trivial_patch)

    gfu = GridFunction(fes)
    gfu.Set(x**3 + sin(y))

    u,v = fes.TnT()
    bf = u*v*dx
    lf = gfu*v*dx

    gfupatch = GridFunction(fes)
    gfupatch.vec.data = PatchwiseSolve(EA,fes,bf,lf)

    Draw(gfu,mesh,"gfuset")
    Draw(gfupatch,mesh,"gfupatch")
    diff = sqrt(Integrate(InnerProduct(gfu-gfupatch,gfu-gfupatch) * dx ,mesh))
    assert(diff < 1e-12)



# ------------------------------ LOAD LIBRARIES -------------------------------
import pytest
from netgen.geom2d import SplineGeometry
from ngsolve import *
from ngsolve.solvers import PreconditionedRichardson as PreRic
from xfem import *
from xfem.mlset import *

ngsglobals.msg_level = 1
SetNumThreads(4)


# -------------------------------- PARAMETERS ---------------------------------
full_order = 4
orders = [1, 2, 3]
h0 = 0.4
mesh_levels = range(0, 6)

rate_tol = 0.125


# ----------------------------------- DATA ------------------------------------
u_ex = 16 * x * (1 - x) * y * (1 - y)
grad_u_ex = CoefficientFunction((16 * (1 - 2 * x) * y * (1 - y),
                                 16 * x * (1 - x) * (1 - 2 * y)))
rhs = 32 * (y * (1 - y) + x * (1 - x))


def level_sets():
    return [-y, x - 1, y - 1, -x]


# --------------------------------- GEOMETRY ----------------------------------
geo = SplineGeometry()
geo.AddRectangle((-0.2, -0.2), (1.2, 1.2),
                 bcs=("bottom", "right", "top", "left"))


# -------------------------------- MAIN RUTINE --------------------------------


def SolvePossionOnUnitSquare(level_sets, mesh, k, rhs, gamma_n=10, gamma_s=0.1, 
        u_ex=None, grad_u_ex=None, inverse="sparsecholesky"):

    nr_ls = len(level_sets)

    # ------------------------- Finite Element Space --------------------------
    V = H1(mesh, order=k, dgjumps=True)

    gfu = GridFunction(V)
    freedofs = BitArray(V.ndof)

    # -------------------------- Levelset & Cut-Info --------------------------
    level_sets_p1 = tuple(GridFunction(H1(mesh, order=1))
                          for i in range(nr_ls))
    for i, lsetp1 in enumerate(level_sets_p1):
        InterpolateToP1(level_sets[i], lsetp1)

    square = DomainTypeArray([(NEG, NEG, NEG, NEG)])
    square.Compress(level_sets_p1)
    lset_dom_inner = {"levelset": level_sets_p1, "domain_type": square}

    boundary = square.Boundary()
    lsets_bnd = {}
    for dtt in boundary:
        lsets_bnd[dtt] = {"levelset": level_sets_p1, "domain_type": dtt}

    mlci = MultiLevelsetCutInfo(mesh, level_sets_p1)

    # ---------------------------- Element Markers ----------------------------
    els_hasneg, els_if = BitArray(mesh.ne), BitArray(mesh.ne)
    els_if_singe = {dtt: BitArray(mesh.ne) for dtt in boundary}
    facets_gp = BitArray(mesh.nedge)

    els_hasneg[:] = False
    els_hasneg |= mlci.GetElementsWithContribution(square)

    els_if[:] = False
    els_if |= els_hasneg & ~mlci.GetElementsOfType(square)

    for i, (dtt, els_bnd) in enumerate(els_if_singe.items()):
        els_bnd[:] = False
        els_bnd |= mlci.GetElementsWithContribution(dtt)

    facets_gp[:] = False
    facets_gp |= GetFacetsWithNeighborTypes(mesh, a=els_hasneg, b=els_if,
                                            use_and=True)

    freedofs[:] = False
    freedofs |= GetDofsOfElements(V, els_hasneg) & V.FreeDofs()

    # --------------------------- (Bi)Linear Forms ----------------------------
    u, v = V.TnT()
    h = specialcf.mesh_size
    normals = square.GetOuterNormals(level_sets_p1)

    diffusion = InnerProduct(Grad(u), Grad(v))

    def nitsche(n):
        return - InnerProduct(Grad(u) * n, v) - InnerProduct(Grad(v) * n, u) \
                    + (gamma_n * k * k / h) * InnerProduct(u, v)
    ghost_penalty = gamma_s / (h**2) * (u - u.Other()) * (v - v.Other())

    forcing = rhs * v

    # ------------------------------ Integrators ------------------------------
    a = RestrictedBilinearForm(V, element_restriction=els_hasneg,
                               facet_restriction=facets_gp, check_unused=False)
    a += SymbolicBFI(lset_dom_inner, form=diffusion,
                     definedonelements=els_hasneg)
    for bnd, n in normals.items():
        a += SymbolicBFI(lsets_bnd[bnd], form=nitsche(n),
                         definedonelements=els_if_singe[bnd])
    a += SymbolicFacetPatchBFI(form=ghost_penalty, skeleton=False,
                               definedonelements=facets_gp)

    f = LinearForm(V)
    f += SymbolicLFI(lset_dom_inner, form=forcing)

    # ----------------------------- Solve Problem -----------------------------
    f.Assemble()
    a.Assemble()
    inv = a.mat.Inverse(freedofs=freedofs,inverse=inverse)

    gfu.vec.data = PreRic(a=a, rhs=f.vec, pre=inv, freedofs=freedofs)

    # --------------------------- Error Computation ---------------------------
    if u_ex:
        err_l2 = sqrt(Integrate(lset_dom_inner, cf=InnerProduct(gfu - u_ex,
                                                                gfu - u_ex),
                                mesh=mesh, order=2 * k))
    else:
        err_l2 = float("NaN")
    if grad_u_ex:
        err_h1 = sqrt(Integrate(lset_dom_inner,
                                cf=InnerProduct(Grad(gfu) - grad_u_ex,
                                                Grad(gfu) - grad_u_ex),
                                mesh=mesh, order=2 * (k - 1)))
    else:
        err_h1 = float("NaN")

    del inv, a, f, gfu
    del els_hasneg, els_if, els_if_singe, facets_gp, freedofs
    del lset_dom_inner, lsets_bnd

    return err_l2, err_h1

# ----------------------------- UTILITY FUNTIONS ------------------------------
def CompLogRate(y1, y2, delta=log(2)):
    return (log(y2) - log(y1)) / delta

# -------------------------------- FIRST CHECK --------------------------------
def test_solution_in_space(): 
    with TaskManager():
        mesh = Mesh(geo.GenerateMesh(maxh=h0))
        errl2, errh1 = SolvePossionOnUnitSquare(
            level_sets(), mesh, full_order, rhs, u_ex=u_ex, grad_u_ex=grad_u_ex)

        assert errl2 < 1e-12, errh1 < 1e-12

        del mesh


# ----------------------------- CONVERGENCE STUDY -----------------------------
def test_check_convergence_order():
    errs = {"l2": {}, "h1": {}}

    with TaskManager():
        for Lx in mesh_levels:

            ngmesh = geo.GenerateMesh(maxh=h0)
            for i in range(Lx):
                ngmesh.Refine()
            mesh = Mesh(ngmesh)

            for k in orders:
                errl2, errh1 = SolvePossionOnUnitSquare(
                    level_sets(), mesh, k, rhs, u_ex=u_ex, grad_u_ex=grad_u_ex,
                    gamma_s=0.5)

                errs["l2"][(Lx, k)] = errl2
                errs["h1"][(Lx, k)] = errh1

            del ngmesh, mesh


    # ---------------------------- Post-Processing ----------------------------
    for k in orders:
        print("\n Results with k = {:}:".format(k))
        print("--------------------------------------------")
        print("h_max   | L2 err.  | Rate | H1 err.  | Rate ")
        print("--------------------------------------------")

        ratesl2, ratesh1 = 0, 0

        for i, Lx in enumerate(mesh_levels):

            if i > 0:
                rate_l2 = CompLogRate(errs["l2"][(Lx, k)],
                                      errs["l2"][(Lx - 1, k)])
                rate_h1 = CompLogRate(errs["h1"][(Lx, k)],
                                      errs["h1"][(Lx - 1, k)])
                if i > 1:
                    ratesl2 += rate_l2 
                    ratesh1 += rate_h1

                print("{:3.1e}   {:5.3e}  {:4.2f}   {:5.3e}  {:4.2f} ".format(
                    h0 * 0.5**Lx, errs["l2"][(Lx, k)], rate_l2,
                    errs["h1"][(Lx, k)], rate_h1))

            else:
                print("{:3.1e}   {:5.3e}  ----   {:5.3e}  ---- ".format(
                    h0 * 0.5**Lx, errs["l2"][(Lx, k)], errs["h1"][(Lx, k)]))
        print("--------------------------------------------")

        ratesl2 = ratesl2 / (len(mesh_levels) - 2)
        ratesh1 = ratesh1 / (len(mesh_levels) - 2)

        print("Average L2 rate (on final meshes): {}".format(ratesl2))
        print("Average H1 rate (on final meshes): {}".format(ratesh1))

        assert k + 1 - ratesl2 < rate_tol
        assert k - ratesh1 < rate_tol


# -----------------------------------------------------------------------------
# --------------------------- RUN TESTS SEPARATELY ----------------------------
# -----------------------------------------------------------------------------
if __name__ == "__main__":
    test_solution_in_space()
    test_check_convergence_order()
